# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os

import pytest

from ansys.fluent.core import FluentVersion
from ansys.fluent.core.examples.downloads import download_file
from ansys.fluent.core.services.datamodel_tui import TUIMenu


@pytest.mark.skip("Failing in github")
def test_report_system_proc_stats_tui(new_solver_session, capsys) -> None:
    new_solver_session.tui.report.system.sys_stats()
    captured = capsys.readouterr()
    assert "CPU" in captured.out


def test_runtime_tui_menus(static_mixer_case_session) -> None:
    solver = static_mixer_case_session
    solver.tui.define.models.addon_module(3)
    rmf = solver.tui.define.models.resolved_MEA_fuelcells
    assert rmf is not None
    assert rmf.__class__ == TUIMenu


@pytest.mark.codegen_required
def test_python_keyword_menu_name(new_meshing_session):
    meshing = new_meshing_session
    assert "cad_options" in dir(meshing.tui.file.import_)
    assert "create_cad_assemblies" in dir(meshing.tui.file.import_.cad_options)
    meshing.tui.file.import_.cad_options.create_cad_assemblies("yes")


@pytest.mark.skip("Failing in github")
def test_api_upgrade_message(new_solver_session):
    solver = new_solver_session
    case_name = download_file(
        "Static_Mixer_main.cas.h5", "pyfluent/static_mixer", save_path=os.getcwd()
    )
    from contextlib import redirect_stdout
    import io

    f = io.StringIO()
    with redirect_stdout(f):
        solver.tui.file.read_case(case_name)
    s = f.getvalue()
    if solver.get_fluent_version() >= FluentVersion.v251:
        assert (
            s.split("\n")[-2].split("(")[0]
            == r"<solver_session>.settings.file.read_case"
        )
    else:
        assert s.split("\n")[-2].split("(")[0] == r"<solver_session>.file.read_case"


def test_exit_not_in_meshing_tui(new_meshing_session):
    meshing = new_meshing_session

    assert "exit" not in dir(meshing.tui)

    with pytest.raises(AttributeError):
        meshing.tui.exit()


def test_commands_not_in_solver_tui(new_solver_session):
    solver = new_solver_session

    deleted_commands = ["exit"]
    hidden_commands = ["switch_to_meshing_mode"]

    for command in deleted_commands:
        assert command not in dir(solver.tui)
        with pytest.raises(AttributeError):
            getattr(solver.tui, command)

    for command in hidden_commands:
        assert command not in dir(solver.tui)
        assert getattr(solver.tui, command)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection and the Base Session."""

from enum import Enum
import json
import logging
from typing import Any, Callable, Dict
import warnings
import weakref

from deprecated.sphinx import deprecated

from ansys.fluent.core.fluent_connection import FluentConnection
from ansys.fluent.core.journaling import Journal
from ansys.fluent.core.launcher.launcher_utils import is_compose
from ansys.fluent.core.pyfluent_warnings import (
    PyFluentDeprecationWarning,
    PyFluentUserWarning,
)
from ansys.fluent.core.services import service_creator
from ansys.fluent.core.services.app_utilities import AppUtilitiesOld
from ansys.fluent.core.services.field_data import FieldDataService, ZoneInfo
from ansys.fluent.core.services.scheme_eval import SchemeEval
from ansys.fluent.core.streaming_services.datamodel_event_streaming import (
    DatamodelEvents,
)
from ansys.fluent.core.streaming_services.events_streaming import EventsManager
from ansys.fluent.core.streaming_services.field_data_streaming import FieldDataStreaming
from ansys.fluent.core.streaming_services.transcript_streaming import Transcript
from ansys.fluent.core.utils.fluent_version import FluentVersion

from .rpvars import RPVars

try:
    from ansys.fluent.core.solver.settings import root
except Exception:
    root = Any

logger = logging.getLogger("pyfluent.general")


def _parse_server_info_file(file_name: str):
    with open(file_name, encoding="utf-8") as f:
        lines = f.readlines()
    ip_and_port = lines[0].strip().split(":")
    ip = ip_and_port[0]
    port = int(ip_and_port[1])
    password = lines[1].strip()
    return ip, port, password


class _IsDataValid:
    def __init__(self, scheme_eval):
        self._scheme_eval = scheme_eval

    def __bool__(self):
        return self()

    def __call__(self):
        return self._scheme_eval.scheme_eval("(data-valid?)")


class _AppUtilitiesFactory:
    """AppUtilities factory."""

    @staticmethod
    def _create_app_utilities(scheme_eval, fluent_connection):
        if FluentVersion(scheme_eval.version) < FluentVersion.v252:
            return AppUtilitiesOld(scheme_eval)
        else:
            return fluent_connection._connection_interface._app_utilities


class BaseSession:
    """Encapsulates a Fluent session.

    This class exposes methods for interacting with a Fluent session.

    Attributes
    ----------
    scheme: SchemeEval
        Instance of ``SchemeEval`` to execute Fluent's scheme code on.

    Methods
    -------
    _create_from_server_info_file(
        server_info_file_name, cleanup_on_exit, start_transcript
        )
        Create a Session instance from server-info file

    exit()
        Close the Fluent connection and exit Fluent.
    """

    # We are passing around an WeakMethod to avoid circular references
    def __init__(
        self,
        fluent_connection: FluentConnection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
        event_type: Enum | None = None,
        get_zones_info: weakref.WeakMethod[Callable[[], list[ZoneInfo]]] | None = None,
    ):
        """BaseSession.

        Parameters
        ----------
        fluent_connection (:ref:`ref_fluent_connection`):
            Encapsulates a Fluent connection.
        scheme_eval: SchemeEval
            Instance of ``SchemeEval`` to execute Fluent's scheme code on.
        file_transfer_service : Optional
            Service for uploading and downloading files.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        event_type : Enum, optional
            Event enumeration specific to the session type.
        """
        self._start_transcript = start_transcript
        self._launcher_args = launcher_args
        BaseSession._build_from_fluent_connection(
            self,
            fluent_connection,
            scheme_eval,
            file_transfer_service,
            event_type,
            get_zones_info,
        )
        self.register_finalizer_callback = fluent_connection.register_finalizer_cb

    def _build_from_fluent_connection(
        self,
        fluent_connection: FluentConnection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        event_type=None,
        get_zones_info: weakref.WeakMethod[Callable[[], list[ZoneInfo]]] | None = None,
    ):
        """Build a BaseSession object from fluent_connection object."""
        self._fluent_connection = fluent_connection
        self._file_transfer_service = file_transfer_service
        self._error_state = fluent_connection._error_state
        self.scheme = scheme_eval
        self.rp_vars = RPVars(self.scheme.string_eval)
        self._preferences = None

        self._transcript_service = service_creator("transcript").create(
            fluent_connection._channel, fluent_connection._metadata
        )
        self.transcript = Transcript(self._transcript_service)
        if self._start_transcript:
            self.transcript.start()

        self._app_utilities = _AppUtilitiesFactory._create_app_utilities(
            self.scheme, self._fluent_connection
        )

        self.journal = Journal(self._app_utilities)

        self._datamodel_service_tui = service_creator("tui").create(
            fluent_connection._channel,
            fluent_connection._metadata,
            self._error_state,
            self._app_utilities,
            self.scheme,
        )

        self._datamodel_service_se = service_creator("datamodel").create(
            fluent_connection._channel,
            fluent_connection._metadata,
            self.get_fluent_version(),
            self._error_state,
            self._file_transfer_service,
        )

        self._datamodel_events = DatamodelEvents(self._datamodel_service_se)
        self._datamodel_events.start()

        self._batch_ops_service = service_creator("batch_ops").create(
            fluent_connection._channel, fluent_connection._metadata
        )

        if event_type:
            events_service = service_creator("events").create(
                fluent_connection._channel, fluent_connection._metadata
            )
            self.events = EventsManager[event_type](
                event_type, events_service, self._error_state, weakref.proxy(self)
            )
            self.events.start()
        else:
            self.events = None

        self._field_data_service = self._fluent_connection.create_grpc_service(
            FieldDataService, self._error_state
        )

        class Fields:
            """Container for field and solution variables."""

            def __init__(self, _session):
                """Initialize Fields."""
                self._is_solution_data_valid = (
                    _session._app_utilities.is_solution_data_available
                )
                self.field_info = service_creator("field_info").create(
                    _session._field_data_service,
                    self._is_solution_data_valid,
                )
                self.field_data = service_creator("field_data").create(
                    _session._field_data_service,
                    self.field_info,
                    self._is_solution_data_valid,
                    _session.scheme,
                    get_zones_info,
                )
                self.field_data_streaming = FieldDataStreaming(
                    _session._fluent_connection._id, _session._field_data_service
                )
                self.field_data_old = service_creator("field_data_old").create(
                    _session._field_data_service,
                    self.field_info,
                    self._is_solution_data_valid,
                    _session.scheme,
                )

        self.fields = Fields(self)

        self._settings_service = service_creator("settings").create(
            fluent_connection._channel,
            fluent_connection._metadata,
            self._app_utilities,
            self.scheme,
            self._error_state,
        )

        self._health_check = fluent_connection._health_check
        self.connection_properties = fluent_connection.connection_properties

        self._fluent_connection.register_finalizer_cb(
            self._datamodel_service_se.unsubscribe_all_events
        )
        for obj in filter(None, (self._datamodel_events, self.transcript, self.events)):
            self._fluent_connection.register_finalizer_cb(obj.stop)

    def is_server_healthy(self) -> bool:
        """Whether the current session is healthy (i.e. The server is 'SERVING')."""
        return self._health_check.is_serving

    def is_active(self) -> bool:
        """Whether the current session is active."""
        return True if self._fluent_connection else False

    @property
    @deprecated(version="0.32", reason="Use ``session.scheme``.")
    def scheme_eval(self):
        """Provides access to Fluent field information."""
        return self.scheme

    @deprecated(version="0.32", reason="Use ``session.is_server_healthy``.")
    def health_check(self):
        """Provides access to Health Check service."""
        return self._health_check

    @property
    @deprecated(version="0.20.dev9", reason="Use ``session.fields.field_info``.")
    def field_info(self):
        """Provides access to Fluent field information."""
        return self.fields.field_info

    @property
    @deprecated(version="0.20.dev9", reason="Use ``session.fields.field_data``.")
    def field_data(self):
        """Fluent field data on surfaces."""
        return self.fields.field_data

    @property
    @deprecated(
        version="0.20.dev9", reason="Use ``session.fields.field_data_streaming``."
    )
    def field_data_streaming(self):
        """Field gRPC streaming service of Fluent."""
        return self.fields.field_data_streaming

    @property
    def id(self) -> str:
        """Return the session ID."""
        return self._fluent_connection._id

    def start_journal(self, file_name: str):
        """Executes tui command to start journal."""
        warnings.warn("Use -> journal.start()", PyFluentDeprecationWarning)
        self.journal.start(file_name)

    def stop_journal(self):
        """Executes tui command to stop journal."""
        warnings.warn("Use -> journal.stop()", PyFluentDeprecationWarning)
        self.journal.stop()

    @classmethod
    def _create_from_server_info_file(
        cls,
        server_info_file_name: str,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
        **connection_kwargs,
    ):
        """Create a Session instance from server-info file.

        Parameters
        ----------
        server_info_file_name : str
            Path to server-info file written out by Fluent server
        file_transfer_service : Optional
            Support file upload and download.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        **connection_kwargs : dict, optional
            Additional keyword arguments may be specified, and they will be passed to the `FluentConnection`
            being initialized. For example, ``cleanup_on_exit = True``.
            See :func:`FluentConnection initialization <ansys.fluent.core.fluent_connection.FluentConnection.__init__>`
            for more details and possible arguments.

        Returns
        -------
        Session
            Session instance
        """
        ip, port, password = _parse_server_info_file(server_info_file_name)
        fluent_connection = FluentConnection(
            ip=ip,
            port=port,
            password=password,
            file_transfer_service=file_transfer_service,
            **connection_kwargs,
        )
        session = cls(
            fluent_connection=fluent_connection,
            scheme_eval=fluent_connection._connection_interface.scheme_eval,
            file_transfer_service=file_transfer_service,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )
        return session

    def execute_tui(self, command: str) -> None:
        """Executes a tui command."""
        self.scheme.eval(f"(ti-menu-load-string {json.dumps(command)})")

    def get_fluent_version(self) -> FluentVersion:
        """Gets and returns the fluent version."""
        return FluentVersion(self.scheme.version)

    def _exit_compose_service(self):
        if self._fluent_connection._container and is_compose():
            self._fluent_connection._container.stop()

    def exit(self, **kwargs) -> None:
        """Exit session."""
        logger.debug("session.exit() called")
        if self._fluent_connection:
            self._exit_compose_service()
            self._fluent_connection.exit(**kwargs)
            self._fluent_connection = None

    def force_exit(self) -> None:
        """Forces the Fluent session to exit, losing unsaved progress and data."""
        self._exit_compose_service()
        self._fluent_connection.force_exit()

    def file_exists_on_remote(self, file_name: str) -> bool:
        """Check if remote file exists.

        Parameters
        ----------
        file_name: str
            File name.

        Returns
        -------
            Whether file exists.
        """
        if self._file_transfer_service:
            return self._file_transfer_service.file_exists_on_remote(file_name)

    def _file_transfer_api_warning(self, method_name: str) -> str:
        """User warning for upload/download methods."""
        return f"You have directly called the {method_name} method of the session. \
        Please be advised that for the current version of Fluent, many API methods \
        automatically handle file uploads and downloads internally. You may not \
        need to explicitly call {method_name} in most cases. \
        However, there are exceptions, particularly in PMFileManagement, where complex \
        file interactions require explicit use of {method_name}  method \
        for relevant files."

    def upload(self, file_name: list[str] | str, remote_file_name: str | None = None):
        """Upload a file to the server.

        Parameters
        ----------
        file_name : str
            Name of the local file to upload to the server.
        remote_file_name : str, optional
            remote file name, by default None
        """
        warnings.warn(self._file_transfer_api_warning("upload()"), PyFluentUserWarning)
        if self._file_transfer_service:
            return self._file_transfer_service.upload(file_name, remote_file_name)

    def download(self, file_name: str, local_directory: str | None = None):
        """Download a file from the server.

        Parameters
        ----------
        file_name : str
            Name of the file to download from the server.
        local_directory : str, optional
            Local destination directory. The default is the current working directory.
        """
        warnings.warn(
            self._file_transfer_api_warning("download()"), PyFluentUserWarning
        )
        if self._file_transfer_service:
            return self._file_transfer_service.download(file_name, local_directory)

    def chdir(self, path: str) -> None:
        """Change Fluent working directory.

        Parameters
        ----------
        path : str
            Path of the directory to change.
        """
        self._app_utilities.set_working_directory(path)

    def __enter__(self):
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any):
        """Close the Fluent connection and exit Fluent."""
        logger.debug("session.__exit__() called")
        self.exit()

    def __dir__(self):
        if self._fluent_connection is None:
            return ["is_active"]
        dir_list = set(list(self.__dict__.keys()) + dir(type(self))) - {
            "field_data",
            "field_info",
            "field_data_streaming",
            "start_journal",
            "stop_journal",
        }
        return sorted(dir_list)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core.examples import download_file, path
from ansys.fluent.core.filereader.casereader import CaseReader


def test_get_and_set_rp_vars(new_solver_session) -> None:
    case_path = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    solver = new_solver_session
    solver.tui.file.read_case(case_path)
    rp_vars = solver.rp_vars

    # simple integer
    iter_count = 54321
    rp_vars("number-of-iterations", iter_count)
    assert iter_count == rp_vars("number-of-iterations")

    # complex list structure
    before_init_mod = rp_vars("strategy/solution-strategy/before-init-modification")
    assert before_init_mod[1][1][1] == ("value", False)
    before_init_mod[1][1][1] = ("value", True)
    rp_vars("strategy/solution-strategy/before-init-modification", before_init_mod)
    before_init_mod_2 = rp_vars("strategy/solution-strategy/before-init-modification")
    assert before_init_mod_2[1][1][1] == ("value", True)


@pytest.mark.fluent_version(">=23.1, !=24.1")
def test_get_all_rp_vars(new_solver_session) -> None:
    case_path = download_file("Static_Mixer_main.cas.h5", "pyfluent/static_mixer")
    solver = new_solver_session
    solver.tui.file.read_case(case_path)
    rp_vars = solver.rp_vars
    # all vars
    all_vars = rp_vars()
    assert len(all_vars) == pytest.approx(9000, 10)

    # refresh
    solver.file.write(file_type="case", file_name=case_path)
    solver.tui.file.read_case(case_path)

    # all vars again
    all_vars = rp_vars()
    assert len(all_vars) == pytest.approx(9000, 20)

    # CaseFile comparison, note that the PyFluent work dir is not necessarily the same as the Fluent work dir
    case = CaseReader(case_file_name=path(case_path))
    case_vars = case.rp_vars()
    assert len(case_vars) == pytest.approx(9000, 450)


@pytest.mark.fluent_version(">=23.2")
def test_rp_vars_allowed_values(new_solver_session) -> None:
    solver = new_solver_session
    rp_vars = solver.rp_vars

    assert rp_vars("number-of-iterations") == 0

    with pytest.raises(RuntimeError):
        rp_vars("number-of-iterat")

    assert "number-of-iterations" in rp_vars.allowed_values()


@pytest.mark.fluent_version(">=23.2")
def test_rp_vars_boolean(new_solver_session) -> None:
    solver = new_solver_session

    var_name = "rp-lam?"
    rp_vars = solver.rp_vars

    var_val = rp_vars(var_name)
    assert isinstance(var_val, bool)

    for i in range(10):
        var_val = not var_val
        rp_vars(var_name, var_val)
        assert rp_vars(var_name) == var_val



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time

import pytest
from util import create_datamodel_root_in_server

from ansys.fluent.core.services.datamodel_se import SubscribeEventError
from ansys.fluent.core.utils.execution import timeout_loop

rule_str = (
    "RULES:\n"
    "  STRING: X\n"
    "    default = ijk\n"
    "  END\n"
    "  SINGLETON: ROOT\n"
    "    members = A, B, D, G\n"
    "    commands= C\n"
    "    SINGLETON: A\n"
    "      members = X\n"
    "      x = $./X\n"
    "    END\n"
    "    OBJECT: B\n"
    "      members = X\n"
    "    END\n"
    "    SINGLETON: D\n"
    "      members = E, F, X\n"
    "      SINGLETON: E\n"
    "        members = X\n"
    "      END\n"
    "      SINGLETON: F\n"
    "        members = X\n"
    "      END\n"
    "    END\n"
    "    SINGLETON: G\n"
    "      members = H\n"
    "      DICT: H\n"
    "      END\n"
    "    END\n"
    "    COMMAND: C\n"
    "      arguments = X\n"
    "      x = $/A/X\n"
    "    END\n"
    "  END\n"
    "END\n"
)


@pytest.mark.fluent_version(">=25.2")
def test_env_var_setting(datamodel_api_version_all, request, new_solver_session):
    solver = new_solver_session
    test_name = request.node.name
    # TODO: It might be possible to check the param value in the fixture
    # instead of checking the test name here.
    if test_name.endswith("[old]"):
        assert solver.scheme.eval('(getenv "REMOTING_NEW_DM_API")') is None
    elif test_name.endswith("[new]"):
        assert solver.scheme.eval('(getenv "REMOTING_NEW_DM_API")') == "1"


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_child_created(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service

    called = 0
    created = []

    def cb(path: str):
        nonlocal called
        nonlocal created
        called += 1
        created.append(path)

    subscription = service.add_on_child_created(app_name, "/", "B", cb)
    assert called == 0
    assert created == []
    service.set_state(app_name, "/", {"B:b": {"_name_": "b"}})
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    assert created == ["/B:b"]
    subscription.unsubscribe()


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_changed(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = 0
    state = None
    called_obj = 0
    state_obj = None

    def cb(value):
        nonlocal called
        nonlocal state
        state = value
        called += 1

    def cb_obj(value):
        nonlocal called_obj
        nonlocal state_obj
        state_obj = value
        called_obj += 1

    subscription = service.add_on_changed(app_name, "/A/X", cb)
    subscription_obj = service.add_on_changed(app_name, "/A", cb_obj)
    assert called == 0
    assert state is None
    assert called_obj == 0
    assert state_obj is None
    service.set_state(app_name, "/A/X", "lmn")
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    assert state == "lmn"
    assert called_obj == 1
    assert state_obj == {"X": "lmn"}
    service.set_state(app_name, "/A/X", "abc")
    timeout_loop(lambda: called == 2, timeout=5)
    assert called == 2
    assert state == "abc"
    assert called_obj == 2
    assert state_obj == {"X": "abc"}
    subscription.unsubscribe()
    subscription_obj.unsubscribe()
    service.set_state(app_name, "/A/X", "xyz")
    time.sleep(5)
    assert called == 2
    assert state == "abc"
    assert called_obj == 2
    assert state_obj == {"X": "abc"}


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_affected(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = 0

    def cb():
        nonlocal called
        called += 1

    subscription = service.add_on_affected(app_name, "/D", cb)
    assert called == 0
    service.set_state(app_name, "/D/X", "lmn")
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    service.set_state(app_name, "/D/E/X", "lmn")
    timeout_loop(lambda: called == 2, timeout=5)
    assert called == 2
    service.set_state(app_name, "/A/X", "lmn")
    time.sleep(5)
    assert called == 2
    subscription.unsubscribe()
    service.set_state(app_name, "/D/E/X", "pqr")
    time.sleep(5)
    assert called == 2


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_affected_at_type_path(
    datamodel_api_version_all, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = 0

    def cb():
        nonlocal called
        called += 1

    subscription = service.add_on_affected_at_type_path(app_name, "/D", "E", cb)
    assert called == 0
    service.set_state(app_name, "/D/X", "lmn")
    time.sleep(5)
    assert called == 0
    service.set_state(app_name, "/D/E/X", "lmn")
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    service.set_state(app_name, "/D/F/X", "lmn")
    time.sleep(5)
    assert called == 1
    subscription.unsubscribe()
    service.set_state(app_name, "/D/E/X", "pqr")
    time.sleep(5)
    assert called == 1


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_deleted(
    datamodel_api_version_all, request, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = False
    called_obj = False

    def cb():
        nonlocal called
        called = True

    def cb_obj():
        nonlocal called_obj
        called_obj = True

    service.set_state(app_name, "/", {"B:b": {"_name_": "b"}})
    subscription = service.add_on_deleted(app_name, "/B:b/X", cb)
    subscription_obj = service.add_on_deleted(app_name, "/B:b", cb_obj)
    assert not called
    assert not called_obj
    service.delete_object(app_name, "/B:b")
    time.sleep(5)
    test_name = request.node.name
    # TODO: Note comment in StateEngine test testDataModelAPIOnDeleted
    if test_name.endswith("[old]"):
        assert called
    elif test_name.endswith("[new]"):
        assert not called
    assert called_obj
    subscription.unsubscribe()
    subscription_obj.unsubscribe()


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_attribute_changed(
    datamodel_api_version_all, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = 0
    value = None

    def cb(val):
        nonlocal called
        nonlocal value
        value = val
        called += 1

    subscription = service.add_on_attribute_changed(app_name, "/A", "x", cb)
    assert called == 0
    assert value is None
    service.set_state(app_name, "/A/X", "cde")
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    assert value == "cde"
    service.set_state(app_name, "/A/X", "xyz")
    timeout_loop(lambda: called == 2, timeout=5)
    assert called == 2
    assert value == "xyz"
    subscription.unsubscribe()
    service.set_state(app_name, "/A/X", "abc")
    time.sleep(5)
    assert called == 2
    assert value == "xyz"


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_command_attribute_changed(
    datamodel_api_version_all, request, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    called = 0
    value = None

    def cb(val):
        nonlocal called
        nonlocal value
        value = val
        called += 1

    subscription = service.add_on_command_attribute_changed(app_name, "/", "C", "x", cb)
    assert called == 0
    assert value is None
    service.set_state(app_name, "/A/X", "cde")
    timeout_loop(lambda: called == 1, timeout=5)
    assert called == 1
    assert value == "cde"
    service.set_state(app_name, "/A/X", "xyz")
    timeout_loop(lambda: called == 2, timeout=5)
    assert called == 2
    test_name = request.node.name
    # TODO: the value is not modiefied in the old API - issue
    if test_name.endswith("[new]"):
        assert value == "xyz"
    subscription.unsubscribe()
    service.set_state(app_name, "/A/X", "abc")
    time.sleep(5)
    assert called == 2
    if test_name.endswith("[new]"):
        assert value == "xyz"


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_command_executed(
    datamodel_api_version_all, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    executed = 0
    command = None
    arguments = None

    def cb(cmd, args):
        nonlocal executed
        nonlocal command
        nonlocal arguments
        command = cmd
        arguments = args
        executed += 1

    subscription = service.add_on_command_executed(app_name, "/", cb)
    assert executed == 0
    assert command is None
    assert arguments is None
    service.execute_command(app_name, "/", "C", dict(X="abc"))
    timeout_loop(lambda: executed == 1, timeout=5)
    assert executed == 1
    assert command == "C"
    assert arguments == {"X": "abc"}
    subscription.unsubscribe()
    service.execute_command(app_name, "/", "C", dict(X="uvw"))
    time.sleep(5)
    assert executed == 1
    assert command == "C"
    assert arguments == {"X": "abc"}


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_get_state(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    assert service.get_state(app_name, "/A/X") == "ijk"


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_set_state(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    service.set_state(app_name, "/A/X", "new_val")
    assert service.get_state(app_name, "/A/X") == "new_val"


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_update_dict(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    service.update_dict(app_name, "/G/H", {"X": "abc"})
    assert service.get_state(app_name, "/G/H") == {"X": "abc"}


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_on_bad_input(
    datamodel_api_version_all, request, new_solver_session
):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    test_name = request.node.name
    new_api = test_name.endswith("[new]")
    with pytest.raises(SubscribeEventError):
        service.add_on_child_created(app_name, "", "", lambda _: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_child_created(app_name, "/BB", "B", lambda _: None)
    with pytest.raises(SubscribeEventError):
        service.add_on_child_created(app_name, "/", "A", lambda _: None)
    with pytest.raises(SubscribeEventError):
        service.add_on_child_created(app_name, "/", "BB", lambda _: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_changed(app_name, "/BB", lambda _: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_deleted(app_name, "/BB", lambda: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_affected(app_name, "/BB", lambda _: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_affected_at_type_path(app_name, "/BB", "B", lambda: None)
    # TODO: not raised in the old API - issue
    if new_api:
        with pytest.raises(SubscribeEventError):
            service.add_on_affected_at_type_path(app_name, "/", "BB", lambda: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_attribute_changed(app_name, "/BB", "isActive", lambda _: None)
    with pytest.raises(SubscribeEventError):
        service.add_on_attribute_changed(app_name, "/A", "", lambda _: None)
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_command_attribute_changed(
            app_name, "/BB", "C", "isActive", lambda _: None
        )
    with pytest.raises(SubscribeEventError):
        service.add_on_command_attribute_changed(
            app_name, "/A", "CC", "", lambda _: None
        )
    with pytest.raises(SubscribeEventError):
        service.add_on_command_attribute_changed(
            app_name, "/", "CC", "isActive", lambda _: None
        )
    with pytest.raises(RuntimeError if new_api else SubscribeEventError):  # TODO: issue
        service.add_on_command_executed(app_name, "/BB", lambda _: None)


@pytest.mark.fluent_version(">=25.2")
def test_datamodel_api_static_info(datamodel_api_version_all, new_solver_session):
    solver = new_solver_session
    app_name = "test"
    create_datamodel_root_in_server(solver, rule_str, app_name)
    service = solver._se_service
    assert service.get_static_info(app_name)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import ast
import importlib
from pathlib import Path
import pickle
import shutil
import tempfile

import pytest

import ansys.fluent.core as pyfluent
from ansys.fluent.core.codegen import StaticInfoType, allapigen
from ansys.fluent.core.codegen.datamodelgen import meshing_rule_file_names
from ansys.fluent.core.search import get_api_tree_file_name
from ansys.fluent.core.utils.fluent_version import get_version_for_file_name


@pytest.mark.codegen_required
def test_allapigen_files(new_solver_session):
    version = get_version_for_file_name(session=new_solver_session)
    importlib.import_module(f"ansys.fluent.core.generated.fluent_version_{version}")
    importlib.import_module(f"ansys.fluent.core.generated.meshing.tui_{version}")
    importlib.import_module(f"ansys.fluent.core.generated.solver.tui_{version}")
    importlib.import_module(f"ansys.fluent.core.generated.datamodel_{version}.meshing")
    importlib.import_module(f"ansys.fluent.core.generated.datamodel_{version}.workflow")
    importlib.import_module(
        f"ansys.fluent.core.generated.datamodel_{version}.preferences"
    )
    importlib.import_module(
        f"ansys.fluent.core.generated.datamodel_{version}.part_management"
    )
    importlib.import_module(
        f"ansys.fluent.core.generated.datamodel_{version}.pm_file_management"
    )
    importlib.import_module(f"ansys.fluent.core.generated.solver.settings_{version}")


def test_codegen_with_no_static_info(monkeypatch):
    codegen_outdir = Path(tempfile.mkdtemp())
    monkeypatch.setattr(pyfluent, "CODEGEN_OUTDIR", codegen_outdir)
    version = "252"
    allapigen.generate(version, {})
    generated_paths = list(codegen_outdir.iterdir())
    assert len(generated_paths) == 1
    assert set(p.name for p in generated_paths) == {f"api_tree_{version}.pickle"}
    api_tree_file = get_api_tree_file_name(version)
    with open(api_tree_file, "rb") as f:
        api_tree = pickle.load(f)
    assert api_tree == {"<meshing_session>": {}, "<solver_session>": {}}


def _get_nth_tui_command_static_info(n):
    return {f"C{n}": {"menus": {}, "commands": {}, "help": f"C{n} help"}}


def _get_nth_tui_menu_static_info(n, menus_static_info, commands_static_info):
    return {
        f"M{n}": {
            "menus": menus_static_info,
            "commands": commands_static_info,
            "help": f"M{n} help",
        }
    }


def _get_expected_tui_api_output(mode):
    return f'''"""Fluent {mode} TUI commands"""
#
# This is an auto-generated file.  DO NOT EDIT!
#
# pylint: disable=line-too-long

from ansys.fluent.core.services.datamodel_tui import PyMenu, TUIMenu, TUIMethod



class main_menu(TUIMenu):
    """
    Fluent {mode} main menu.
    """
    def __init__(self, service, version, mode, path):
        self.M1 = self.__class__.M1(service, version, mode, path + ["M1"])
        self.C1 = self.__class__.C1(service, version, mode, path + ["C1"])
        super().__init__(service, version, mode, path)
    class C1(TUIMethod):
        """
        C1 help.
        """

    class M1(TUIMenu):
        """
        M1 help.
        """
        def __init__(self, service, version, mode, path):
            self.M2 = self.__class__.M2(service, version, mode, path + ["M2"])
            self.C2 = self.__class__.C2(service, version, mode, path + ["C2"])
            super().__init__(service, version, mode, path)
        class C2(TUIMethod):
            """
            C2 help.
            """

        class M2(TUIMenu):
            """
            M2 help.
            """
            def __init__(self, service, version, mode, path):
                self.C3 = self.__class__.C3(service, version, mode, path + ["C3"])
                super().__init__(service, version, mode, path)
            class C3(TUIMethod):
                """
                C3 help.
                """'''


@pytest.mark.parametrize("mode", ["solver", "meshing"])
def test_codegen_with_tui_solver_static_info(mode, monkeypatch):
    codegen_outdir = Path(tempfile.mkdtemp())
    monkeypatch.setattr(pyfluent, "CODEGEN_OUTDIR", codegen_outdir)
    version = "252"
    static_infos = {}
    static_info_type = (
        StaticInfoType.TUI_SOLVER if mode == "solver" else StaticInfoType.TUI_MESHING
    )
    static_infos[static_info_type] = {
        "menus": _get_nth_tui_menu_static_info(
            1,
            _get_nth_tui_menu_static_info(2, {}, _get_nth_tui_command_static_info(3)),
            _get_nth_tui_command_static_info(2),
        ),
        "commands": _get_nth_tui_command_static_info(1),
        "help": "Root",
    }
    allapigen.generate(version, static_infos)
    generated_paths = list(codegen_outdir.iterdir())
    assert len(generated_paths) == 2
    assert set(p.name for p in generated_paths) == {f"api_tree_{version}.pickle", mode}
    solver_paths = list((codegen_outdir / mode).iterdir())
    assert len(solver_paths) == 1
    assert set(p.name for p in solver_paths) == {f"tui_{version}.py"}
    with open(codegen_outdir / mode / f"tui_{version}.py", "r") as f:
        assert f.read().strip() == _get_expected_tui_api_output(mode)
    api_tree_file = get_api_tree_file_name(version)
    with open(api_tree_file, "rb") as f:
        api_tree = pickle.load(f)
    tui_tree = {
        "tui": {"M1": {"M2": {"C3": "Command"}, "C2": "Command"}, "C1": "Command"}
    }
    api_tree_expected = {"<meshing_session>": {}, "<solver_session>": {}}
    api_tree_expected[f"<{mode}_session>"] = tui_tree
    assert api_tree == api_tree_expected
    shutil.rmtree(str(codegen_outdir))


_static_info_type_by_rules = {
    "workflow": StaticInfoType.DATAMODEL_WORKFLOW,
    "meshing": StaticInfoType.DATAMODEL_MESHING,
    "PartManagement": StaticInfoType.DATAMODEL_PART_MANAGEMENT,
    "PMFileManagement": StaticInfoType.DATAMODEL_PM_FILE_MANAGEMENT,
    "flicing": StaticInfoType.DATAMODEL_FLICING,
    "preferences": StaticInfoType.DATAMODEL_PREFERENCES,
    "solverworkflow": StaticInfoType.DATAMODEL_SOLVER_WORKFLOW,
    "MeshingUtilities": StaticInfoType.DATAMODEL_MESHING_UTILITIES,
}


def _get_datamodel_entity_static_info(
    name, singletons, namedobjects, commands, parameters
):
    return {
        name: {
            "singletons": singletons,
            "namedobjects": namedobjects,
            "commands": commands,
            "parameters": parameters,
        }
    }


def _get_datamodel_command_static_info(name, args, returntype):
    args = [{"name": arg[0], "type": arg[1]} for arg in args]
    return {name: {"commandinfo": {"args": args, "returntype": returntype}}}


_expected_datamodel_api_output = '''#
# This is an auto-generated file.  DO NOT EDIT!
#
# pylint: disable=line-too-long

from ansys.fluent.core.services.datamodel_se import (
    PyMenu,
    PyParameter,
    PyTextual,
    PyNumerical,
    PyDictionary,
    PyNamedObjectContainer,
    PyCommand,
    PyQuery,
    PyCommandArguments,
    PyTextualCommandArgumentsSubItem,
    PyNumericalCommandArgumentsSubItem,
    PyDictionaryCommandArgumentsSubItem,
    PyParameterCommandArgumentsSubItem,
    PySingletonCommandArgumentsSubItem
)


class Root(PyMenu):
    """
    Singleton Root.
    """
    def __init__(self, service, rules, path):
        self.N1 = self.__class__.N1(service, rules, path + [("N1", "")])
        self.S1 = self.__class__.S1(service, rules, path + [("S1", "")])
        self.P1 = self.__class__.P1(service, rules, path + [("P1", "")])
        self.C1 = self.__class__.C1(service, rules, "C1", path)
        super().__init__(service, rules, path)

    class N1(PyNamedObjectContainer):
        """
        .
        """
        class _N1(PyMenu):
            """
            Singleton _N1.
            """
            def __init__(self, service, rules, path):
                self.S3 = self.__class__.S3(service, rules, path + [("S3", "")])
                super().__init__(service, rules, path)

            class S3(PyMenu):
                """
                Singleton S3.
                """
                def __init__(self, service, rules, path):
                    super().__init__(service, rules, path)

        def __getitem__(self, key: str) -> _N1:
            return super().__getitem__(key)

    class S1(PyMenu):
        """
        Singleton S1.
        """
        def __init__(self, service, rules, path):
            self.S2 = self.__class__.S2(service, rules, path + [("S2", "")])
            self.P2 = self.__class__.P2(service, rules, path + [("P2", "")])
            self.C2 = self.__class__.C2(service, rules, "C2", path)
            super().__init__(service, rules, path)

        class S2(PyMenu):
            """
            Singleton S2.
            """
            def __init__(self, service, rules, path):
                super().__init__(service, rules, path)

        class P2(PyNumerical):
            """
            Parameter P2 of value type float.
            """
            pass

        class C2(PyCommand):
            """
            Command C2.

            Parameters
            ----------
            A2 : float

            Returns
            -------
            bool
            """
            class _C2CommandArguments(PyCommandArguments):
                def __init__(self, service, rules, command, path, id):
                    super().__init__(service, rules, command, path, id)
                    self.A2 = self._A2(self, "A2", service, rules, path)

                class _A2(PyNumericalCommandArgumentsSubItem):
                    """
                    Argument A2.
                    """

            def create_instance(self) -> _C2CommandArguments:
                args = self._get_create_instance_args()
                if args is not None:
                    return self._C2CommandArguments(*args)

    class P1(PyTextual):
        """
        Parameter P1 of value type str.
        """
        pass

    class C1(PyCommand):
        """
        Command C1.

        Parameters
        ----------
        A1 : str

        Returns
        -------
        bool
        """
        class _C1CommandArguments(PyCommandArguments):
            def __init__(self, service, rules, command, path, id):
                super().__init__(service, rules, command, path, id)
                self.A1 = self._A1(self, "A1", service, rules, path)

            class _A1(PyTextualCommandArgumentsSubItem):
                """
                Argument A1.
                """

        def create_instance(self) -> _C1CommandArguments:
            args = self._get_create_instance_args()
            if args is not None:
                return self._C1CommandArguments(*args)'''


@pytest.mark.parametrize(
    "rules",
    [
        "workflow",
        "meshing",
        "PartManagement",
        "PMFileManagement",
        "preferences",
        "solverworkflow",
        "MeshingUtilities",
    ],
)
def test_codegen_with_datamodel_static_info(monkeypatch, rules):
    codegen_outdir = Path(tempfile.mkdtemp())
    monkeypatch.setattr(pyfluent, "CODEGEN_OUTDIR", codegen_outdir)
    version = "251"
    static_infos = {}
    static_info_type = _static_info_type_by_rules[rules]
    static_infos[static_info_type] = {
        "singletons": _get_datamodel_entity_static_info(
            "S1",
            _get_datamodel_entity_static_info("S2", {}, {}, {}, {}),
            {},
            _get_datamodel_command_static_info("C2", [("A2", "Real")], "Logical"),
            {"P2": {"type": "Real"}},
        ),
        "namedobjects": _get_datamodel_entity_static_info(
            "N1", _get_datamodel_entity_static_info("S3", {}, {}, {}, {}), {}, {}, {}
        ),
        "commands": _get_datamodel_command_static_info(
            "C1", [("A1", "String")], "Logical"
        ),
        "parameters": {"P1": {"type": "String"}},
        "help": "Root",
    }
    allapigen.generate(version, static_infos)
    generated_paths = list(codegen_outdir.iterdir())
    assert len(generated_paths) == 2
    assert set(p.name for p in generated_paths) == {
        f"api_tree_{version}.pickle",
        f"datamodel_{version}",
    }
    datamodel_paths = list((codegen_outdir / f"datamodel_{version}").iterdir())
    assert len(datamodel_paths) == 1 or 2
    assert set(p.name for p in datamodel_paths) == {
        f"{meshing_rule_file_names[rules]}.py"
    } or {f"{meshing_rule_file_names[rules]}.pyi"}
    with open(
        codegen_outdir
        / f"datamodel_{version}"
        / f"{meshing_rule_file_names[rules]}.py",
        "r",
    ) as f:
        assert f.read().strip() == _expected_datamodel_api_output
    api_tree_file = get_api_tree_file_name(version)
    with open(api_tree_file, "rb") as f:
        api_tree = pickle.load(f)
    datamodel_tree = {
        rules: {
            "C1": "Command",
            "N1:<name>": {"S3": {}},
            "P1": "Parameter",
            "S1": {"C2": "Command", "P2": "Parameter", "S2": {}},
        }
    }
    api_tree_expected = {"<meshing_session>": {}, "<solver_session>": {}}
    if rules in [
        "workflow",
        "meshing",
        "PartManagement",
        "PMFileManagement",
        "preferences",
        "MeshingUtilities",
    ]:
        api_tree_expected["<meshing_session>"] = datamodel_tree
    if rules in ["workflow", "flicing", "preferences", "solverworkflow"]:
        api_tree_expected["<solver_session>"] = datamodel_tree
    assert api_tree == api_tree_expected
    shutil.rmtree(str(codegen_outdir))


def _get_group_settings_static_info(name, children, commands, queries):
    return {
        name: {
            "children": children,
            "commands": commands,
            "queries": queries,
            "type": "group",
            "help": f"{name} help",
        }
    }


def _get_named_object_settings_static_info(name, object_type, children):
    return {
        name: {
            "object-type": object_type,
            "children": children,
            "type": "named-object",
            "help": f"{name} help",
        }
    }


def _get_parameter_settings_static_info(name, type_):
    return {name: {"type": type_, "help": f"{name} help"}}


def _get_command_settings_static_info(name, args):
    args = {arg[0]: {"type": arg[1], "help": f"{arg[0]} help"} for arg in args}
    return {name: {"arguments": args, "type": "command", "help": f"{name} help"}}


def _get_query_settings_static_info(name, args):
    args = {arg[0]: {"type": arg[1], "help": f"{arg[0]} help"} for arg in args}
    return {name: {"arguments": args, "type": "query", "help": f"{name} help"}}


_settings_static_info = {
    "children": (
        _get_group_settings_static_info(
            "G1",
            (
                (
                    _get_group_settings_static_info(
                        "G2",
                        _get_parameter_settings_static_info("P3", "integer"),
                        {},
                        {},
                    )
                )
                | _get_parameter_settings_static_info("P2", "real")
            ),
            _get_command_settings_static_info("C2", [("A2", "real")]),
            _get_query_settings_static_info("Q2", [("A2", "real")]),
        )
        | _get_parameter_settings_static_info("P1", "string")
        | _get_named_object_settings_static_info(
            "N1",
            _get_group_settings_static_info("G3", {}, {}, {})["G3"],
            _get_parameter_settings_static_info("P4", "string"),
        )
    ),
    "commands": _get_command_settings_static_info("C1", [("A1", "string")]),
    "queries": _get_query_settings_static_info("Q1", [("A1", "string")]),
    "type": "group",
}


_expected_settings_api_output = '''#
# This is an auto-generated file.  DO NOT EDIT!
#

from ansys.fluent.core.solver.flobject import *

from ansys.fluent.core.solver.flobject import (
    _ChildNamedObjectAccessorMixin,
    _NonCreatableNamedObjectMixin,
    _InputFile,
    _OutputFile,
    _InOutFile,
)

SHASH = "3e6d76a4601701388ea8258912d145b7b7c436699a50b6c7fe9a29f41eeff194"

class P3(Integer):
    """
    P3 help.
    """
    _version = '251'
    fluent_name = 'P3'
    _python_name = 'P3'

class G2(Group):
    """
    G2 help.
    """
    _version = '251'
    fluent_name = 'G2'
    _python_name = 'G2'
    child_names = ['P3']
    _child_classes = dict(
        P3=P3,
    )

class P2(Real):
    """
    P2 help.
    """
    _version = '251'
    fluent_name = 'P2'
    _python_name = 'P2'

class A2(Real):
    """
    A2 help.
    """
    _version = '251'
    fluent_name = 'A2'
    _python_name = 'A2'

class C2(Command):
    """
    C2 help.
    
    Parameters
    ----------
        A2 : real
            A2 help.
    """
    _version = '251'
    fluent_name = 'C2'
    _python_name = 'C2'
    argument_names = ['A2']
    _child_classes = dict(
        A2=A2,
    )

class Q2(Query):
    """
    Q2 help.
    
    Parameters
    ----------
        A2 : real
            A2 help.
    """
    _version = '251'
    fluent_name = 'Q2'
    _python_name = 'Q2'
    argument_names = ['A2']
    _child_classes = dict(
        A2=A2,
    )

class G1(Group):
    """
    G1 help.
    """
    _version = '251'
    fluent_name = 'G1'
    _python_name = 'G1'
    child_names = ['G2', 'P2']
    command_names = ['C2']
    query_names = ['Q2']
    _child_classes = dict(
        G2=G2,
        P2=P2,
        C2=C2,
        Q2=Q2,
    )

class P1(String):
    """
    P1 help.
    """
    _version = '251'
    fluent_name = 'P1'
    _python_name = 'P1'

class P4(String):
    """
    P4 help.
    """
    _version = '251'
    fluent_name = 'P4'
    _python_name = 'P4'

class N1_child(Group):
    """
    'child_object_type' of N1.
    """
    _version = '251'
    fluent_name = 'child-object-type'
    _python_name = 'N1_child'

class N1(NamedObject[N1_child], _NonCreatableNamedObjectMixin[N1_child]):
    """
    N1 help.
    """
    _version = '251'
    fluent_name = 'N1'
    _python_name = 'N1'
    child_names = ['P4']
    _child_classes = dict(
        P4=P4,
    )
    child_object_type = N1_child

class A1(String):
    """
    A1 help.
    """
    _version = '251'
    fluent_name = 'A1'
    _python_name = 'A1'

class C1(Command):
    """
    C1 help.
    
    Parameters
    ----------
        A1 : str
            A1 help.
    """
    _version = '251'
    fluent_name = 'C1'
    _python_name = 'C1'
    argument_names = ['A1']
    _child_classes = dict(
        A1=A1,
    )

class Q1(Query):
    """
    Q1 help.
    
    Parameters
    ----------
        A1 : str
            A1 help.
    """
    _version = '251'
    fluent_name = 'Q1'
    _python_name = 'Q1'
    argument_names = ['A1']
    _child_classes = dict(
        A1=A1,
    )

class root(Group):
    """
    'root' object.
    """
    _version = '251'
    fluent_name = ''
    _python_name = 'root'
    child_names = ['G1', 'P1', 'N1']
    command_names = ['C1']
    query_names = ['Q1']
    _child_classes = dict(
        G1=G1,
        P1=P1,
        N1=N1,
        C1=C1,
        Q1=Q1,
    )'''  # noqa: W293


def test_codegen_with_settings_static_info(monkeypatch):
    codegen_outdir = Path(tempfile.mkdtemp())
    monkeypatch.setattr(pyfluent, "CODEGEN_OUTDIR", codegen_outdir)
    version = "251"
    static_infos = {}
    static_infos[StaticInfoType.SETTINGS] = _settings_static_info
    allapigen.generate(version, static_infos)
    generated_paths = list(codegen_outdir.iterdir())
    assert len(generated_paths) == 2
    assert set(p.name for p in generated_paths) == {
        f"api_tree_{version}.pickle",
        "solver",
    }
    assert {p.name for p in (codegen_outdir / "solver").glob("*.py*")} == {
        f"settings_{version}.py",
        f"settings_{version}.pyi",
    }
    with open(codegen_outdir / "solver" / f"settings_{version}.py", "r") as f:
        assert f.read().strip() == _expected_settings_api_output
    api_tree_file = get_api_tree_file_name(version)
    with open(api_tree_file, "rb") as f:
        api_tree = pickle.load(f)
    settings_tree = {
        "C1": "Command",
        "G1": {
            "C2": "Command",
            "G2": {"P3": "Parameter"},
            "P2": "Parameter",
            "Q2": "Query",
        },
        "N1:<name>": {"P4": "Parameter"},
        "P1": "Parameter",
        "Q1": "Query",
    }
    api_tree_expected = {}
    api_tree_expected["<meshing_session>"] = {}
    api_tree_expected["<solver_session>"] = settings_tree
    assert api_tree == api_tree_expected
    shutil.rmtree(str(codegen_outdir))


_settings_static_info_duplicate_parameters = {
    "children": (
        _get_group_settings_static_info(
            "G1",
            (
                (
                    _get_group_settings_static_info(
                        "G2",
                        _get_parameter_settings_static_info("P1", "string"),
                        {},
                        {},
                    )
                )
                | _get_parameter_settings_static_info("P1", "string")
            ),
            {},
            {},
        )
    ),
    "commands": {},
    "queries": {},
    "type": "group",
}

_settings_static_info_different_parameters_with_same_name = {
    "children": (
        _get_group_settings_static_info(
            "G1",
            (
                (
                    _get_group_settings_static_info(
                        "G2",
                        _get_parameter_settings_static_info("P1", "real"),
                        {},
                        {},
                    )
                )
                | _get_parameter_settings_static_info("P1", "string")
            ),
            {},
            {},
        )
    ),
    "commands": {},
    "queries": {},
    "type": "group",
}


_settings_static_info_combined_case = {
    "children": (
        _get_group_settings_static_info(
            "G1",
            (
                (
                    _get_group_settings_static_info(
                        "G2",
                        (
                            _get_parameter_settings_static_info("P1", "real")
                            | _get_parameter_settings_static_info("P2", "string")
                        ),
                        {},
                        {},
                    )
                )
                | _get_parameter_settings_static_info("P1", "string")
                | _get_parameter_settings_static_info("P2", "string")
            ),
            {},
            {},
        )
    ),
    "commands": {},
    "queries": {},
    "type": "group",
}


@pytest.mark.parametrize(
    "settings_static_info,class_names",
    [
        (_settings_static_info_duplicate_parameters, ["P1", "G2", "G1", "root"]),
        (
            _settings_static_info_different_parameters_with_same_name,
            ["P1_1", "G2", "P1", "G1", "root"],
        ),
        (_settings_static_info_combined_case, ["P1_1", "P2", "G2", "P1", "G1", "root"]),
    ],
)
def test_codegen_with_settings_static_info_edge_cases(
    monkeypatch, settings_static_info, class_names
):

    codegen_outdir = Path(tempfile.mkdtemp())
    monkeypatch.setattr(pyfluent, "CODEGEN_OUTDIR", codegen_outdir)
    version = "251"
    static_infos = {}
    static_infos[StaticInfoType.SETTINGS] = settings_static_info
    allapigen.generate(version, static_infos)
    with open(codegen_outdir / "solver" / f"settings_{version}.py", "r") as f:
        module_def = ast.parse(f.read())
        class_names_from_file = [
            x.name for x in module_def.body if isinstance(x, ast.ClassDef)
        ]
        # The order of classes is important.
        assert class_names_from_file == class_names
    shutil.rmtree(str(codegen_outdir))



"""A script to print the name of the hanging test by parsing the GitHub logs file."""

import argparse


def strip_before_tests(input_string):
    """
    Strips everything before the 'tests/' in the given string.
    Parameters
    ----------
    input_string: str
        The original string to process.
    Returns
    -------
        The modified string.
    """
    index = input_string.find("tests/")
    if index != -1:
        return input_string[index:]


def parse_github_log(log_file_path: str) -> str:
    """
    Parse GitHub log lines to identify hanging tests.

    Parameters
    ----------
    log_file_path: str
        The path of the log file from GitHub.

    Returns
    -------
        Name of the haging tests.
    """
    tests = {}
    hanging_tests = []

    with open(log_file_path, "r", encoding="utf-8") as f:
        log_lines = f.readlines()

    for line in log_lines:
        test_name = strip_before_tests(line)
        if test_name:
            if tests.get(test_name):
                tests[test_name] += 1
            else:
                tests[test_name] = 1

    for key, value in tests.items():
        if value != 2:
            hanging_tests.append(key)

    return hanging_tests


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="A script to print the name of the hanging test by parsing the GitHub logs file."
    )
    parser.add_argument(
        "-logfile",
        type=str,
        help="Path of the log file.",
    )
    args = parser.parse_args()
    hanging_tests = parse_github_log(args.logfile)
    if hanging_tests:
        print("\n Hanging tests detected.\n")
        for test in hanging_tests:
            print(test)
    else:
        print("No hanging tests detected.")



#
# This is an auto-generated file.  DO NOT EDIT!
#
# pylint: disable=line-too-long

from ansys.fluent.core.services.datamodel_se import PyMenu
from typing import Any


class Root(PyMenu):

    def add_labels_on_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def add_labels_on_edge_zones(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def add_labels_on_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def clean_face_zone_names(
        self,
        ) -> None: ...

    def delete_all_sub_domains(
        self,
        ) -> None: ...

    def delete_empty_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def delete_empty_edge_zones(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def delete_empty_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def delete_empty_zones(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def delete_marked_faces_in_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def merge_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> None: ...

    def merge_cell_zones_with_same_prefix(
        self,
        prefix: str | None = None,
        ) -> None: ...

    def merge_cell_zones_with_same_suffix(
        self,
        suffix: str | None = None,
        ) -> None: ...

    def merge_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> None: ...

    def merge_face_zones_of_type(
        self,
        face_zone_type: str | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> None: ...

    def merge_face_zones_with_same_prefix(
        self,
        prefix: str | None = None,
        ) -> None: ...

    def remove_id_suffix_from_face_zones(
        self,
        ) -> None: ...

    def remove_ids_from_zone_names(
        self,
        zone_id_list: list[int] | None = None,
        ) -> bool: ...

    def remove_labels_on_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> None: ...

    def remove_labels_on_edge_zones(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> None: ...

    def remove_labels_on_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> None: ...

    def rename_edge_zone(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        new_name: str | None = None,
        ) -> bool: ...

    def rename_face_zone(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        new_name: str | None = None,
        ) -> bool: ...

    def rename_face_zone_label(
        self,
        object_name: str | None = None,
        old_label_name: str | None = None,
        new_label_name: str | None = None,
        ) -> None: ...

    def rename_object(
        self,
        old_object_name: str | None = None,
        new_object_name: str | None = None,
        ) -> None: ...

    def renumber_zone_ids(
        self,
        zone_id_list: list[int] | None = None,
        start_number: int | None = None,
        ) -> None: ...

    def replace_cell_zone_suffix(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        old_suffix: str | None = None,
        new_suffix: str | None = None,
        merge: bool | None = None,
        ) -> None: ...

    def replace_edge_zone_suffix(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        old_suffix: str | None = None,
        new_suffix: str | None = None,
        merge: bool | None = None,
        ) -> None: ...

    def replace_face_zone_suffix(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        separator: str | None = None,
        replace_with: str | None = None,
        merge: bool | None = None,
        ) -> None: ...

    def replace_label_suffix(
        self,
        object_name_list: list[str] | None = None,
        separator: str | None = None,
        new_suffix: str | None = None,
        ) -> None: ...

    def replace_object_suffix(
        self,
        object_name_list: list[str] | None = None,
        separator: str | None = None,
        new_suffix: str | None = None,
        ) -> None: ...

    def set_number_of_parallel_compute_threads(
        self,
        nthreads: int | None = None,
        ) -> None: ...

    def set_object_cell_zone_type(
        self,
        object_name: str | None = None,
        cell_zone_type: str | None = None,
        ) -> None: ...

    def set_quality_measure(
        self,
        measure: str | None = None,
        ) -> None: ...

    def _cell_zones_labels_fdl(
        self,
        ) -> None: ...

    def _cell_zones_str_fdl(
        self,
        ) -> None: ...

    def _edge_zones_labels_fdl(
        self,
        ) -> None: ...

    def _edge_zones_str_fdl(
        self,
        ) -> None: ...

    def _face_zones_labels_fdl(
        self,
        ) -> None: ...

    def _face_zones_str_fdl(
        self,
        ) -> None: ...

    def _node_zones_labels_fdl(
        self,
        ) -> None: ...

    def _node_zones_str_fdl(
        self,
        ) -> None: ...

    def _object_names_str_fdl(
        self,
        ) -> None: ...

    def _prism_cell_zones_labels_fdl(
        self,
        ) -> None: ...

    def _prism_cell_zones_str_fdl(
        self,
        ) -> None: ...

    def _regions_str_fdl(
        self,
        ) -> None: ...

    def _zone_types_fdl(
        self,
        ) -> None: ...

    def boundary_zone_exists(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        ) -> bool: ...

    def cell_zone_exists(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        ) -> bool: ...

    def convert_zone_ids_to_name_strings(
        self,
        zone_id_list: list[int] | None = None,
        ) -> list[str]: ...

    def convert_zone_name_strings_to_ids(
        self,
        zone_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def copy_face_zone_labels(
        self,
        from_face_zone_id: int | None = None,
        from_face_zone_name: str | None = None,
        to_face_zone_id: int | None = None,
        to_face_zone_name: str | None = None,
        ) -> None: ...

    def count_marked_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def create_boi_and_size_functions_from_refinement_regions(
        self,
        region_type: str | None = None,
        boi_prefix_string: str | None = None,
        create_size_function: bool | None = None,
        ) -> None: ...

    def dump_face_zone_orientation_in_region(
        self,
        file_name: str | None = None,
        ) -> None: ...

    def fill_holes_in_face_zone_list(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        max_hole_edges: int | None = None,
        ) -> None: ...

    def get_adjacent_cell_zones_for_given_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_adjacent_face_zones_for_given_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_adjacent_interior_and_boundary_face_zones_for_given_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_adjacent_zones_by_edge_connectivity(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_adjacent_zones_by_node_connectivity(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_all_objects(
        self,
        ) -> list[str]: ...

    def get_average_bounding_box_center(
        self,
        face_zone_id_list: list[int] | None = None,
        ) -> list[float]: ...

    def get_baffles_for_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        ) -> list[int]: ...

    def get_bounding_box_of_zone_list(
        self,
        zone_id_list: list[int] | None = None,
        ) -> None: ...

    def get_cell_mesh_distribution(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        measure: str | None = None,
        partitions: int | None = None,
        range: list[float] | None = None,
        ) -> None: ...

    def get_cell_quality_limits(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        measure: str | None = None,
        ) -> None: ...

    def get_cell_zone_count(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> int: ...

    def get_cell_zone_id_list_with_labels(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def get_cell_zone_shape(
        self,
        cell_zone_id: int | None = None,
        ) -> str: ...

    def get_cell_zone_volume(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_cell_zones(
        self,
        maximum_entity_count: float | None = None,
        xyz_coordinates: list[float] | None = None,
        filter: str | None = None,
        ) -> list[int]: ...

    def get_edge_size_limits(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[float]: ...

    def get_edge_zone_id_list_with_labels(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def get_edge_zones(
        self,
        maximum_entity_count: float | None = None,
        only_boundary: bool | None = None,
        filter: str | None = None,
        ) -> list[int]: ...

    def get_edge_zones_list(
        self,
        filter: list[str] | None = None,
        ) -> None: ...

    def get_edge_zones_of_object(
        self,
        objects: list[str] | None = None,
        object_name: str | None = None,
        ) -> list[int]: ...

    def get_embedded_baffles(
        self,
        ) -> list[int]: ...

    def get_face_mesh_distribution(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        measure: str | None = None,
        partitions: int | None = None,
        range: list[float] | None = None,
        ) -> None: ...

    def get_face_quality_limits(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        measure: str | None = None,
        ) -> None: ...

    def get_face_zone_area(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_face_zone_count(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def get_face_zone_id_list_with_labels(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        label_name_list: list[str] | None = None,
        ) -> list[int]: ...

    def get_face_zone_node_count(
        self,
        face_zone_id: int | None = None,
        face_zone_name: str | None = None,
        ) -> int: ...

    def get_face_zones(
        self,
        maximum_entity_count: float | None = None,
        only_boundary: bool | None = None,
        prism_control_name: str | None = None,
        xyz_coordinates: list[float] | None = None,
        filter: str | None = None,
        ) -> list[int]: ...

    def get_face_zones_by_zone_area(
        self,
        maximum_zone_area: float | None = None,
        minimum_zone_area: float | None = None,
        ) -> list[int]: ...

    def get_face_zones_of_object(
        self,
        regions: list[str] | None = None,
        labels: list[str] | None = None,
        region_type: str | None = None,
        objects: list[str] | None = None,
        object_name: str | None = None,
        ) -> list[int]: ...

    def get_face_zones_with_zone_specific_prisms_applied(
        self,
        ) -> list[int]: ...

    def get_free_faces_count(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def get_interior_face_zones_for_given_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_labels(
        self,
        object_name: str | None = None,
        filter: str | None = None,
        label_name_pattern: str | None = None,
        ) -> list[str]: ...

    def get_labels_on_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[str]: ...

    def get_labels_on_edge_zones(
        self,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        ) -> list[str]: ...

    def get_labels_on_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[str]: ...

    def get_labels_on_face_zones_list(
        self,
        face_zone_id_list: list[int] | None = None,
        ) -> None: ...

    def get_maxsize_cell_zone_by_count(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_maxsize_cell_zone_by_volume(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_minsize_face_zone_by_area(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_minsize_face_zone_by_count(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> float: ...

    def get_multi_faces_count(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def get_node_zones(
        self,
        filter: str | None = None,
        ) -> list[int]: ...

    def get_objects(
        self,
        type_name: str | None = None,
        filter: str | None = None,
        ) -> list[str]: ...

    def get_overlapping_face_zones(
        self,
        face_zone_name_pattern: str | None = None,
        area_tolerance: float | None = None,
        distance_tolerance: float | None = None,
        ) -> list[int]: ...

    def get_pairs_of_overlapping_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        join_tolerance: float | None = None,
        absolute_tolerance: bool | None = None,
        join_angle: float | None = None,
        ) -> list[int]: ...

    def get_prism_cell_zones(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_region_volume(
        self,
        object_name: str | None = None,
        region_name: str | None = None,
        sorting_order: str | None = None,
        ) -> None: ...

    def get_regions(
        self,
        object_name: str | None = None,
        region_name_pattern: str | None = None,
        filter: str | None = None,
        ) -> list[str]: ...

    def get_regions_of_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[str]: ...

    def get_shared_boundary_face_zones_for_given_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_tet_cell_zones(
        self,
        zone_id_list: list[int] | None = None,
        zone_name_list: list[str] | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_unreferenced_cell_zones(
        self,
        filter: str | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_unreferenced_edge_zones(
        self,
        filter: str | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_unreferenced_face_zones(
        self,
        filter: str | None = None,
        zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_wrapped_face_zones(
        self,
        ) -> list[int]: ...

    def get_zone_type(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        ) -> str: ...

    def get_zones(
        self,
        type_name: str | None = None,
        group_name: str | None = None,
        ) -> list[int]: ...

    def get_zones_with_free_faces_for_given_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_zones_with_marked_faces_for_given_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def get_zones_with_multi_faces_for_given_face_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> list[int]: ...

    def interior_zone_exists(
        self,
        zone_id: int | None = None,
        zone_name: str | None = None,
        ) -> bool: ...

    def mark_bad_quality_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        quality_limit: float | None = None,
        number_of_rings: int | None = None,
        ) -> int: ...

    def mark_duplicate_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def mark_face_strips_by_height_and_quality(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        strip_type: int | None = None,
        strip_height: float | None = None,
        quality_measure: str | None = None,
        quality_limit: float | None = None,
        feature_angle: float | None = None,
        ) -> int: ...

    def mark_faces_by_quality(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        quality_measure: str | None = None,
        quality_limit: float | None = None,
        append_marking: bool | None = None,
        ) -> None: ...

    def mark_faces_deviating_from_size_field(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        min_size_factor: float | None = None,
        max_size_factor: float | None = None,
        size_factor_type_to_compare: str | None = None,
        ) -> int: ...

    def mark_faces_in_self_proximity(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        relative_tolerance: bool | None = None,
        tolerance: float | None = None,
        proximity_angle: float | None = None,
        ignore_orientation: bool | None = None,
        ) -> None: ...

    def mark_faces_using_node_degree(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        node_degree_threshold: int | None = None,
        ) -> None: ...

    def mark_free_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def mark_invalid_normals(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> int: ...

    def mark_island_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        island_face_count: int | None = None,
        ) -> None: ...

    def mark_multi_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        fringe_length: int | None = None,
        ) -> int: ...

    def mark_point_contacts(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> None: ...

    def mark_self_intersecting_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        mark_folded: bool | None = None,
        ) -> int: ...

    def mark_sliver_faces(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        max_height: float | None = None,
        skew_limit: float | None = None,
        ) -> int: ...

    def mark_spikes(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        spike_angle: float | None = None,
        ) -> int: ...

    def mark_steps(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        step_angle: float | None = None,
        step_width: float | None = None,
        ) -> int: ...

    def mesh_check(
        self,
        type_name: str | None = None,
        edge_zone_id_list: list[int] | None = None,
        edge_zone_name_list: list[str] | None = None,
        edge_zone_name_pattern: str | None = None,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> None: ...

    def mesh_exists(
        self,
        ) -> None: ...

    def print_worst_quality_cell(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        measure: str | None = None,
        ) -> None: ...

    def project_zone_on_plane(
        self,
        zone_id: int | None = None,
        plane: dict[str, Any] | None = None,
        ) -> None: ...

    def refine_marked_faces_in_zones(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> None: ...

    def scale_cell_zones_around_pivot(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        scale: list[float] | None = None,
        pivot: list[float] | None = None,
        use_bbox_center: bool | None = None,
        ) -> None: ...

    def scale_face_zones_around_pivot(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        scale: list[float] | None = None,
        pivot: list[float] | None = None,
        use_bbox_center: bool | None = None,
        ) -> None: ...

    def separate_cell_zone_layers_by_face_zone_using_id(
        self,
        cell_zone_id: int | None = None,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        nlayers: int | None = None,
        ) -> None: ...

    def separate_cell_zone_layers_by_face_zone_using_name(
        self,
        cell_zone_name: str | None = None,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        nlayers: int | None = None,
        ) -> None: ...

    def separate_face_zones_by_cell_neighbor(
        self,
        face_zone_id_list: list[int] | None = None,
        face_zone_name_list: list[str] | None = None,
        face_zone_name_pattern: str | None = None,
        ) -> None: ...

    def unpreserve_cell_zones(
        self,
        cell_zone_id_list: list[int] | None = None,
        cell_zone_name_list: list[str] | None = None,
        cell_zone_name_pattern: str | None = None,
        ) -> None: ...



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from ansys.fluent.core.examples.downloads import download_file


@pytest.mark.settings_only
def test_initialization_settings(new_solver_session):
    solver = new_solver_session
    case_name = download_file(
        "wigley.cas.h5",
        "pyfluent/wigley_hull",
    )
    solver.file.read(
        file_type="case",
        file_name=case_name,
        lightweight_setup=True,
    )
    solver.parallel.partition.set.laplace_smoothing.enabled = True
    solver.parallel.partition.method(partition_method="metis", count=2)
    copy_by_name = solver.setup.materials.database.copy_by_name
    copy_by_name(type="fluid", name="air")
    copy_by_name(type="fluid", name="water-liquid")
    solver.setup.models.multiphase.models = "vof"
    solver.setup.general.operating_conditions.gravity = {
        "enable": True,
        "components": [0.0, 0.0, -9.81],
    }
    solver.setup.general.solver.time = "steady"

    solver.tui.define.models.multiphase.vof_sub_models("yes", "no")
    solver.tui.define.phases.set_domain_properties.change_phases_names("water", "air")
    solver.setup.boundary_conditions.pressure_inlet["inflow"].phase["mixture"] = {
        "multiphase": {
            "open_channel": True,
            "vmag": 1.452,
            "ht_bottom": -0.941875,
        },
        "momentum": {"direction_specification_method": "Direction Vector"},
        "turbulence": {
            "turbulent_intensity": 0.01,
            "turbulent_viscosity_ratio": 1,
        },
    }
    solver.setup.boundary_conditions.pressure_outlet["outflow"].phase["mixture"] = {
        "multiphase": {
            "open_channel": True,
            "ht_bottom": -0.941875,
            "den_spec": "From Free Surface Level",
        },
        "momentum": {
            "direction_spec": "Normal to Boundary",
            "p_backflow_spec_gen": "Static Pressure",
        },
        "turbulence": {
            "turbulent_intensity": 0.01,
            "turbulent_viscosity_ratio": 1,
        },
    }

    solver.solution.methods.p_v_coupling.flow_scheme = "Coupled"
    solver.solution.methods.p_v_coupling.coupled_form = True
    solver.solution.controls.advanced.multi_grid.amg_controls.coupled_parameters.coarsening_parameters.laplace_coarsening = (
        True
    )
    solver.solution.initialization.open_channel_auto_init = {
        "boundary_zone": 3,
        "flat_init": True,
    }
    assert solver.solution.initialization.open_channel_auto_init() == {
        "boundary_zone": 3,
        "flat_init": True,
    }


@pytest.mark.fluent_version(">=24.1")
def test_fmg_initialize(new_solver_session):
    solver = new_solver_session
    case_name = download_file("vki_turbine.cas.gz", "pyfluent/vki_turbine")
    solver.file.read(file_type="case", file_name=case_name)
    solver.mesh.check()
    solver.solution.initialization.standard_initialize()
    solver.solution.initialization.fmg.fmg_initialize()
    solver.tui.solve.iterate(2)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a module for Fluent connection functionality."""

from __future__ import annotations

import ctypes
from ctypes import c_int, sizeof
from dataclasses import dataclass
import itertools
import logging
import os
from pathlib import Path
import platform
import socket
import subprocess
import threading
from typing import Any, Callable, List, Tuple, TypeVar
import weakref

from deprecated.sphinx import deprecated
import grpc

import ansys.fluent.core as pyfluent
from ansys.fluent.core.launcher.launcher_utils import is_compose
from ansys.fluent.core.services import service_creator
from ansys.fluent.core.services.app_utilities import (
    AppUtilitiesOld,
    AppUtilitiesService,
)
from ansys.fluent.core.services.scheme_eval import SchemeEvalService
from ansys.fluent.core.utils.execution import timeout_exec, timeout_loop
from ansys.fluent.core.utils.file_transfer_service import ContainerFileTransferStrategy
from ansys.platform.instancemanagement import Instance

logger = logging.getLogger("pyfluent.general")


def _docker():
    import docker

    return docker


class PortNotProvided(ValueError):
    """Raised when port is not provided."""

    def __init__(self):
        """Initialize PortNotProvided."""
        super().__init__(
            "Provide the 'port' to connect to an existing Fluent instance."
        )


class UnsupportedRemoteFluentInstance(ValueError):
    """Raised when 'wait_process_finished' does not support remote Fluent session."""

    def __init__(self):
        """Initialize UnsupportedRemoteFluentInstance."""
        super().__init__("Remote Fluent instance is unsupported.")


class WaitTypeError(TypeError):
    """Raised when invalid ``wait`` type is provided."""

    def __init__(self):
        """Initialize WaitTypeError."""
        super().__init__("Invalid 'wait' type.")


def _get_max_c_int_limit() -> int:
    """Get the maximum limit of a C int.

    Returns
    -------
    int
        The maximum limit of a C int
    """
    return 2 ** (sizeof(c_int) * 8 - 1) - 1


class MonitorThread(threading.Thread):
    """A class used for monitoring a Fluent session.

    Daemon thread which will ensure cleanup of session objects, shutdown of
    non-deamon threads etc.

    Attributes
    ----------
    cbs : List[Callable]
        Cleanup/shutdown functions
    """

    def __init__(self):
        """Initialize MonitorThread."""
        super().__init__(daemon=True)
        self.cbs: List[Callable] = []

    def run(self) -> None:
        """Run monitor thread."""
        main_thread = threading.main_thread()
        main_thread.join()
        for cb in self.cbs:
            cb()


ContainerT = TypeVar("ContainerT")


def get_container(container_id_or_name: str) -> bool | ContainerT | None:
    """Get the Docker container object.

    Returns
    -------
    bool | Container | None
        If the system is not correctly set up to run Docker containers, returns ``None``.
        If the container was not found, returns ``False``.
        If the container is found, returns the associated Docker container object.

    Notes
    -----
    See `Docker container`_ for more information.

    .. _Docker container: https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.Container
    """
    if not isinstance(container_id_or_name, str):
        container_id_or_name = str(container_id_or_name)
    try:
        docker_client = _docker().from_env()
        container = docker_client.containers.get(container_id_or_name)
    except _docker().errors.NotFound:  # NotFound is a child from DockerException
        return False
    # DockerException is the most common exception we can get here.
    # However, in some system setup, we get ReadTimeoutError from urllib3 library
    # (https://github.com/ansys/pyfluent/issues/3425).
    # As urllib3 is not a direct dependency of PyFluent, we don't want to import it here,
    # hence we catch the generic Exception.
    except Exception as exc:
        logger.info(f"{type(exc).__name__}: {exc}")
        return None
    return container


class ErrorState:
    """Object to indicate the error state of the connected Fluent client.

    Examples
    --------
    >>> import ansys.fluent.core as pyfluent
    >>> session = pyfluent.launch_fluent()
    >>> session._fluent_connection._error_state.set("test", "test details")
    >>> session._fluent_connection._error_state.name
    'test'
    >>> session._fluent_connection._error_state.details
    'test details'
    >>> session._fluent_connection._error_state.clear()
    >>> session._fluent_connection._error_state.name
    ''
    """

    @property
    def name(self):
        """Get name."""
        return self._name

    @property
    def details(self):
        """Get details."""
        return self._details

    def __init__(self, name: str = "", details: str | None = ""):
        """Initializes the error state object.

        Parameters
        ----------
            name : str
                The name of the error state, by default an empty string, indicating no errors.
            details : str
                Additional details of the error, by default an empty string.
        """
        self._name = name
        self._details = details

    def set(self, name: str, details: str):
        """Method to set the error state name and details to new values."""
        self._name = name
        self._details = details

    def clear(self):
        """Method to clear the current error state, emptying the error name and details
        properties."""
        self._name = ""
        self._details = ""


@dataclass(frozen=True)
class FluentConnectionProperties:
    """Stores Fluent connection properties, including connection IP, port and password;
    Fluent Cortex working directory, process ID and hostname; and whether Fluent was
    launched in a docker container.

    Examples
    --------
    These properties are also available through the session object and can be accessed as:

    >>> import ansys.fluent.core as pyfluent
    >>> session = pyfluent.launch_fluent()
    >>> session.connection_properties.list_names()
    ['ip', 'port', 'password', 'cortex_pwd', 'cortex_pid', 'cortex_host', 'inside_container']
    >>> session.connection_properties.ip
    '127.0.0.1'
    """

    ip: str | None = None
    port: int | None = None
    password: str | None = None
    cortex_pwd: str | None = None
    cortex_pid: int | None = None
    cortex_host: str | None = None
    fluent_host_pid: int | None = None
    inside_container: bool | ContainerT | None = None

    def list_names(self) -> list:
        """Returns list with all property names."""
        return [k for k, _ in vars(self).items()]

    def list_values(self) -> dict:
        """Returns dictionary with all property names and values."""
        return vars(self)


def _get_ip_and_port(ip: str | None = None, port: int | None = None) -> (str, int):
    if not ip:
        ip = os.getenv("PYFLUENT_FLUENT_IP", "127.0.0.1")
    if not port:
        port = os.getenv("PYFLUENT_FLUENT_PORT")
    if not port:
        raise PortNotProvided()
    return ip, port


def _get_channel(ip: str, port: int):
    # Same maximum message length is used in the server
    max_message_length = _get_max_c_int_limit()
    return grpc.insecure_channel(
        f"{ip}:{port}",
        options=[
            ("grpc.max_send_message_length", max_message_length),
            ("grpc.max_receive_message_length", max_message_length),
        ],
    )


class _ConnectionInterface:
    def __init__(self, create_grpc_service, error_state):
        self._scheme_eval_service = create_grpc_service(SchemeEvalService, error_state)
        self.scheme_eval = service_creator("scheme_eval").create(
            self._scheme_eval_service
        )
        if (
            pyfluent.FluentVersion(self.scheme_eval.version)
            < pyfluent.FluentVersion.v252
        ):
            self._app_utilities = AppUtilitiesOld(self.scheme_eval)
        else:
            self._app_utilities_service = create_grpc_service(
                AppUtilitiesService, error_state
            )
            self._app_utilities = service_creator("app_utilities").create(
                self._app_utilities_service
            )

    @property
    def product_build_info(self) -> str:
        """Get Fluent build information."""
        build_info = self._app_utilities.get_build_info()
        return f'Build Time: {build_info["build_time"]}  Build Id: {build_info["build_id"]}  Revision: {build_info["vcs_revision"]}  Branch: {build_info["vcs_branch"]}'

    def get_cortex_connection_properties(self):
        """Get connection properties of Fluent."""
        from grpc._channel import _InactiveRpcError

        try:
            logger.info(self.product_build_info)
            logger.debug("Obtaining Cortex connection properties...")
            cortex_info = self._app_utilities.get_controller_process_info()
            solver_info = self._app_utilities.get_solver_process_info()
            fluent_host_pid = solver_info["process_id"]
            cortex_host = cortex_info["hostname"]
            cortex_pid = cortex_info["process_id"]
            cortex_pwd = cortex_info["working_directory"]
            logger.debug("Cortex connection properties successfully obtained.")
        except _InactiveRpcError:
            logger.warning(
                "Fluent Cortex properties unobtainable. 'force exit()' and other "
                "methods are not going to work properly. Proceeding..."
            )
            fluent_host_pid = None
            cortex_host = None
            cortex_pid = None
            cortex_pwd = None

        return fluent_host_pid, cortex_host, cortex_pid, cortex_pwd

    def get_mode(self):
        """Get the mode of a running fluent session."""
        return self._app_utilities.get_app_mode()

    def exit_server(self):
        """Exits the server."""
        self._app_utilities.exit()


def _pid_exists(pid):
    if platform.system() == "Linux":
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True
    elif platform.system() == "Windows":
        process_query_limited_information = 0x1000
        process_handle = ctypes.windll.kernel32.OpenProcess(
            process_query_limited_information, 0, pid
        )
        if process_handle == 0:
            return False
        else:
            ctypes.windll.kernel32.CloseHandle(process_handle)
            return True


class FluentConnection:
    """Encapsulates a Fluent connection.

    Methods
    -------
    exit()
        Close the Fluent connection and exit Fluent.
    """

    _on_exit_cbs: List[Callable] = []
    _id_iter = itertools.count()
    _monitor_thread: MonitorThread | None = None

    def __init__(
        self,
        ip: str | None = None,
        port: int | None = None,
        password: str | None = None,
        channel: grpc.Channel | None = None,
        cleanup_on_exit: bool = True,
        remote_instance: Instance | None = None,
        file_transfer_service: Any | None = None,
        slurm_job_id: str | None = None,
        inside_container: bool | None = None,
        container: ContainerT | None = None,
    ):
        """Initialize a Session.

        Parameters
        ----------
        ip : str, optional
            IP address to connect to existing Fluent instance. Used only
            when ``channel`` is ``None``.  Defaults to ``"127.0.0.1"``
            and can also be set by the environment variable
            ``PYFLUENT_FLUENT_IP=<ip>``.
        port : int, optional
            Port to connect to existing Fluent instance. Used only
            when ``channel`` is ``None``.  Defaults value can be set by
            the environment variable ``PYFLUENT_FLUENT_PORT=<port>``.
        password : str, optional
            Password to connect to existing Fluent instance.
        channel : grpc.Channel, optional
            Grpc channel to use to connect to existing Fluent instance.
            ip and port arguments will be ignored when channel is
            specified.
        cleanup_on_exit : bool, optional
            When True, the connected Fluent session will be shut down
            when PyFluent is exited or exit() is called on the session
            instance, by default True.
        remote_instance : ansys.platform.instancemanagement.Instance
            The corresponding remote instance when Fluent is launched through
            PyPIM. This instance will be deleted when calling
            ``Session.exit()``.
        file_transfer_service : optional
            File transfer service for uploading files to and
            downloading files from the server.
        slurm_job_id: bool, optional
            Job ID of a Fluent session running within a Slurm environment.
        inside_container: bool, optional
            Whether the Fluent session that is being connected to
            is running inside a Docker container.
        container: ContainerT, optional
            The container instance if the Fluent session is running inside
            a container.

        Raises
        ------
        PortNotProvided
            If port is not provided.
        """
        self._error_state = ErrorState()
        self._data_valid = False
        self._channel_str = None
        self._slurm_job_id = None
        self.finalizer_cbs = []
        if channel is not None:
            self._channel = channel
        else:
            ip, port = _get_ip_and_port(ip, port)
            self._channel = _get_channel(ip, port)
            self._channel_str = f"{ip}:{port}"
        self._metadata: List[Tuple[str, str]] = (
            [("password", password)] if password else []
        )

        self._health_check = service_creator("health_check").create(
            self._channel, self._metadata, self._error_state
        )
        # At this point, the server must be running. If the following check_health()
        # throws, we should not proceed.
        # TODO: Show user-friendly error message.
        if pyfluent.CHECK_HEALTH:
            self._health_check.check_health()

        self._slurm_job_id = slurm_job_id

        self._id = f"session-{next(FluentConnection._id_iter)}"

        if not FluentConnection._monitor_thread:
            FluentConnection._monitor_thread = MonitorThread()
            FluentConnection._monitor_thread.start()

        self._connection_interface = _ConnectionInterface(
            self.create_grpc_service, self._error_state
        )
        fluent_host_pid, cortex_host, cortex_pid, cortex_pwd = (
            self._connection_interface.get_cortex_connection_properties()
        )
        self._cleanup_on_exit = cleanup_on_exit
        self._container = container
        if (
            (inside_container is None or inside_container is True)
            and not remote_instance
            and cortex_host is not None
        ):
            logger.info("Checking if Fluent is running inside a container.")
            if not is_compose():
                inside_container = get_container(cortex_host)
                logger.debug(f"get_container({cortex_host}): {inside_container}")
            if inside_container is False:
                logger.info("Fluent is not running inside a container.")
            elif inside_container is None:
                logger.info(
                    "The current system does not support Docker containers. "
                    "Assuming Fluent is not inside a container."
                )

        self.connection_properties = FluentConnectionProperties(
            ip,
            port,
            password,
            cortex_pwd,
            cortex_pid,
            cortex_host,
            fluent_host_pid,
            inside_container,
        )

        self._remote_instance = remote_instance

        self._file_transfer_service = file_transfer_service

        self._exit_event = threading.Event()

        # session.exit() is handled in the daemon thread (MonitorThread) which ensures
        # shutdown of non-daemon threads. A daemon thread is terminated abruptly
        # during interpreter exit, after all non-daemon threads are exited.
        # self._waiting_thread is a long-running thread which is exited
        # at the end of session.exit() to ensure everything within session.exit()
        # gets executed during exit.
        self._waiting_thread = threading.Thread(target=self._exit_event.wait)
        self._waiting_thread.start()

        self._finalizer = weakref.finalize(
            self,
            FluentConnection._exit,
            self._channel,
            self._cleanup_on_exit,
            self._connection_interface,
            self.finalizer_cbs,
            self._remote_instance,
            self._file_transfer_service,
            self._exit_event,
        )
        FluentConnection._monitor_thread.cbs.append(self._finalizer)

    @property
    @deprecated(version="0.32", reason="No longer required at this level.")
    def health_check(self):
        """Provides access to Health Check service."""
        return self._health_check

    def _close_slurm(self):
        subprocess.run(["scancel", f"{self._slurm_job_id}"])

    def force_exit(self):
        """Immediately terminates the Fluent client, losing unsaved progress and data.

        Notes
        -----
        If the Fluent session is responsive, prefer using :func:`exit()` instead.

        Raises
        ------
        subprocess.CalledProcessError
            If the cleanup script fails to execute.

        Examples
        --------
        >>> import ansys.fluent.core as pyfluent
        >>> session = pyfluent.launch_fluent()
        >>> session.force_exit()
        """
        if self.connection_properties.inside_container or is_compose():
            self._force_exit_container()
        elif self._remote_instance is not None:
            logger.error("Cannot execute cleanup script, Fluent running remotely.")
            return
        else:
            pwd = self.connection_properties.cortex_pwd
            pid = self.connection_properties.fluent_host_pid
            host = self.connection_properties.cortex_host
            if host != socket.gethostname():
                logger.error(
                    "Fluent host is not the current host, cancelling forced exit..."
                )
                return
            if os.name == "nt":
                cleanup_file_ext = "bat"
                cmd_list = []
            elif os.name == "posix":
                cleanup_file_ext = "sh"
                cmd_list = ["bash"]
            else:
                logger.error(
                    "Unrecognized or unsupported operating system, cancelling Fluent cleanup script execution."
                )
                return
            cleanup_file_name = f"cleanup-fluent-{host}-{pid}.{cleanup_file_ext}"
            logger.debug(f"Looking for {cleanup_file_name}...")
            cleanup_file_name = Path(pwd, cleanup_file_name)
            if cleanup_file_name.is_file():
                logger.info(
                    f"Executing Fluent cleanup script, file path: {cleanup_file_name}"
                )
                cmd_list.append(cleanup_file_name)
                logger.debug(f"Cleanup command list = {cmd_list}")
                result = subprocess.run(
                    cmd_list,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    timeout=120,
                    check=True,
                )
                if result.returncode != 0:
                    raise subprocess.CalledProcessError(result.returncode, cmd_list)
            elif self._slurm_job_id:
                logger.debug("Fluent running inside Slurm, closing Slurm session...")
                self._close_slurm()
            else:
                logger.error("Could not find cleanup file.")

    def _force_exit_container(self):
        """Immediately terminates the Fluent client running inside a container, losing
        unsaved progress and data."""
        if is_compose() and self._container:
            self._container.stop()
        else:
            container = self.connection_properties.inside_container
            container_id = self.connection_properties.cortex_host
            pid = self.connection_properties.fluent_host_pid
            cleanup_file_name = f"cleanup-fluent-{container_id}-{pid}.sh"
            logger.debug(
                f"Executing Fluent container cleanup script: {cleanup_file_name}"
            )
            if get_container(container_id):
                try:
                    container.exec_run(["bash", cleanup_file_name], detach=True)
                except _docker().errors.APIError as e:
                    logger.info(f"{type(e).__name__}: {e}")
                    logger.debug(
                        "Caught Docker APIError, Docker container probably not running anymore."
                    )
            else:
                logger.debug(
                    "Container not found, cancelling cleanup script execution."
                )

    def register_finalizer_cb(self, cb, at_start=False):
        """Register a callback to run with the finalizer."""
        if at_start:
            self.finalizer_cbs.insert(0, cb)
        else:
            self.finalizer_cbs.append(cb)

    def create_grpc_service(self, service, *args):
        """Create a gRPC service.

        Parameters
        ----------
        service : Any
            service class
        args : Any, optional
            additional arguments, by default empty

        Returns
        -------
        Any
            service object
        """
        return service(self._channel, self._metadata, *args)

    def wait_process_finished(self, wait: float | int | bool = 60):
        """Returns ``True`` if local Fluent processes have finished, ``False`` if they
        are still running when wait limit (default 60 seconds) is reached. Immediately
        cancels and returns ``None`` if ``wait`` is set to ``False``.

        Parameters
        ----------
        wait : float, int or bool, optional
            How long to wait for processes to finish before returning, by default 60 seconds.
            Can also be set to ``True``, which will result in waiting indefinitely.

        Raises
        ------
        UnsupportedRemoteFluentInstance
            If current Fluent instance is running remotely.
        WaitTypeError
            If ``wait`` is specified improperly.
        """
        if self._remote_instance:
            raise UnsupportedRemoteFluentInstance()
        if isinstance(wait, bool):
            if wait:
                wait = 60
            else:
                logger.debug("Wait limit set to 'False', cancelling process wait.")
                return
        if isinstance(wait, (float, int)):
            logger.info(f"Waiting {wait} seconds for Fluent processes to finish...")
        else:
            raise WaitTypeError()

        if self.connection_properties.inside_container and not is_compose():
            _response = timeout_loop(
                get_container,
                wait,
                args=(self.connection_properties.cortex_host,),
                idle_period=0.5,
                expected="falsy",
            )
        else:
            _response = timeout_loop(
                lambda connection: _pid_exists(connection.fluent_host_pid)
                or _pid_exists(connection.cortex_pid),
                wait,
                args=(self.connection_properties,),
                idle_period=0.5,
                expected="falsy",
            )
        return not _response

    def exit(
        self,
        timeout: float | None = None,
        timeout_force: bool = True,
        wait: float | int | bool | None = False,
    ) -> None:
        """Close the Fluent connection and exit Fluent.

        Parameters
        ----------
        timeout : float, optional
            Time in seconds before considering that the exit request has timed out.
            If omitted or specified as None, then the request will not time out and will lock up the interpreter
            while waiting for a response. Will return earlier if request succeeds earlier.
        timeout_force : bool, optional
            If not specified, defaults to ``True``. If ``True``, attempts to terminate the Fluent process if
            exit request reached timeout. If no timeout is set, this option is ignored.
            Executes :func:`force_exit()` or :func:`force_exit_container()`,
            depending on how Fluent was launched.
        wait : float, int or bool, optional
            Specifies whether to wait for local Fluent processes to finish completely before proceeding.
            If omitted or specified as ``False``, will proceed as usual without
            waiting for the Fluent processes to finish.
            Can be set to ``True`` which will wait for up to 60 seconds,
            or set to a float or int value to specify the wait limit.
            If wait limit is reached, will forcefully terminate the Fluent process.
            If set to wait, will return as soon as processes completely finish.
            Does not work for remote Fluent processes.

        Notes
        -----
        Can also set the ``PYFLUENT_TIMEOUT_FORCE_EXIT`` environment variable to specify the number of seconds and
        alter the default ``timeout`` value. Setting this env var to a non-number value, such as ``OFF``,
        will return this function to default behavior. Note that the environment variable will be ignored if
        timeout is specified when calling this function.

        Examples
        --------

        >>> import ansys.fluent.core as pyfluent
        >>> session = pyfluent.launch_fluent()
        >>> session.exit()
        """

        if wait is not False and self._remote_instance:
            logger.warning(
                "Session exit 'wait' option is ignored when working with remote Fluent sessions."
            )

        if timeout is None:
            env_timeout = os.getenv("PYFLUENT_TIMEOUT_FORCE_EXIT")

            if env_timeout:
                logger.debug("Found PYFLUENT_TIMEOUT_FORCE_EXIT env var")
                try:
                    timeout = float(env_timeout)
                    logger.debug(f"Setting TIMEOUT_FORCE_EXIT to {timeout}")
                except ValueError:
                    logger.debug(
                        "Off or unrecognized PYFLUENT_TIMEOUT_FORCE_EXIT value, not enabling timeout force exit"
                    )

        if timeout is None:
            logger.info("Finalizing Fluent connection...")
            self._finalizer()
            if wait is not False:
                self.wait_process_finished(wait=wait)
        else:
            if not timeout_exec(lambda: self._health_check.is_serving, 5):
                logger.debug("gRPC service not working, cancelling soft exit call.")
            else:
                logger.info("Attempting to send exit request to Fluent...")
                success = timeout_exec(self._finalizer, timeout)
                if success:
                    if wait is not False:
                        if self.wait_process_finished(wait=wait):
                            return
                    else:
                        return

            logger.debug("Continuing...")
            if (timeout is not None and timeout_force) or isinstance(
                wait, (float, int)
            ):
                if self._remote_instance:
                    logger.warning("Cannot force exit from Fluent remote instance.")
                    return
                elif self.connection_properties.inside_container:
                    logger.debug(
                        "Fluent running inside container, cleaning up Fluent inside container..."
                    )
                    self.force_exit()
                else:
                    logger.debug(
                        "Fluent running locally, cleaning up Fluent processes..."
                    )
                    self.force_exit()
                logger.debug("Done.")
            else:
                logger.debug("Timeout and wait force exit disabled, returning...")

    @staticmethod
    def _exit(
        channel,
        cleanup_on_exit,
        connection_interface,
        finalizer_cbs,
        remote_instance,
        file_transfer_service,
        exit_event,
    ) -> None:
        logger.debug("FluentConnection exit method called.")
        if channel:
            for cb in finalizer_cbs:
                cb()
            if cleanup_on_exit:
                try:
                    connection_interface.exit_server()
                except Exception:
                    pass
            channel.close()
            channel = None

        if remote_instance:
            remote_instance.delete()

        if file_transfer_service and isinstance(
            file_transfer_service, ContainerFileTransferStrategy
        ):
            file_transfer_service.container.kill()

        exit_event.set()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".. _doe_ml:

Design of Experiments and Machine Learning model building
---------------------------------------------------------
"""

#######################################################################################
# Objective
# =====================================================================================
#
# Water enters a Mixing Elbow from two Inlets; Hot (313 K) and Cold (293 K) and exits
# from Outlet. Using PyFluent in the background, this example runs Design of Experiments
# with Cold Inlet Velocity and Hot Inlet Velocity as Input Parameters and Outlet
# Temperature as an Output Parameter.
#
# Results can be visualized using a Response Surface. Finally, Supervised Machine
# Learning Regression Task is performed to build the ML Model.
#
# This example demonstrates:
#
# * Design of Experiment, Fluent setup and simulation using PyFluent.
# * Building of Supervised Machine Learning Model.

###################################
# Import required libraries/modules
# =================================

# flake8: noqa: E402

from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import seaborn as sns
import tensorflow as tf
from tensorflow import keras

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

###########################################################################
# Specifying save path
# ====================
# * save_path can be specified as Path("E:/", "pyfluent-examples-tests") or
# * Path("E:/pyfluent-examples-tests") in a Windows machine for example,  or
# * Path("~/pyfluent-examples-tests") in Linux.

save_path = Path(pyfluent.EXAMPLES_PATH)

import_filename = examples.download_file(
    "elbow.cas.h5",
    "pyfluent/examples/DOE-ML-Mixing-Elbow",
    save_path=save_path,
)

#######################
# Fluent Solution Setup
# =====================

#################################################################
# Launch Fluent session with solver mode and print Fluent version
# ===============================================================

solver = pyfluent.launch_fluent(
    precision="double",
    processor_count=2,
    version="3d",
)
print(solver.get_fluent_version())


#############################################################################
# Read case
# =========

solver.settings.file.read_case(file_name=import_filename)

##############################################################################################
# Design of Experiments
# =====================
# * Define Manual DOE as numpy arrays
# * Run cases in sequence
# * Populate results (Mass Weighted Average of Temperature at Outlet) in resArr

coldVelArr = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7])
hotVelArr = np.array([0.8, 1, 1.2, 1.4, 1.6, 1.8, 2.0])
resArr = np.zeros((coldVelArr.shape[0], hotVelArr.shape[0]))

for idx1, coldVel in np.ndenumerate(coldVelArr):
    for idx2, hotVel in np.ndenumerate(hotVelArr):
        cold_inlet = solver.settings.setup.boundary_conditions.velocity_inlet[
            "cold-inlet"
        ]
        cold_inlet.momentum.velocity.value = coldVel

        hot_inlet = solver.settings.setup.boundary_conditions.velocity_inlet[
            "hot-inlet"
        ]
        hot_inlet.momentum.velocity.value = hotVel

        solver.settings.solution.initialization.initialization_type = "standard"
        solver.settings.solution.initialization.standard_initialize()
        solver.settings.solution.run_calculation.iterate(iter_count=200)

        res_tui = solver.scheme.exec(
            (
                "(ti-menu-load-string "
                '"/report/surface-integrals/mass-weighted-avg outlet () '
                'temperature no")',
            )
        )
        resArr[idx1][idx2] = eval(res_tui.split(" ")[-1])


##############################################################################################
# Close the session
# =================

solver.exit()

####################################
# Plot Response Surface using Plotly
# ==================================

fig = go.Figure(data=[go.Surface(z=resArr.T, x=coldVelArr, y=hotVelArr)])

fig.update_layout(
    title={
        "text": "Mixing Elbow Response Surface",
        "y": 0.9,
        "x": 0.5,
        "xanchor": "center",
        "yanchor": "top",
    }
)

fig.update_layout(
    scene=dict(
        xaxis_title="Cold Inlet Vel (m/s)",
        yaxis_title="Hot Inlet Vel (m/s)",
        zaxis_title="Outlet Temperature (K)",
    ),
    width=600,
    height=600,
    margin=dict(l=80, r=80, b=80, t=80),
)
fig.show()


#####################################
# Supervised ML for a Regression Task
# ===================================

############################################
# Create Pandas Dataframe for ML Model Input
# ==========================================
coldVelList = []
hotVelList = []
ResultList = []

for idx1, coldVel in np.ndenumerate(coldVelArr):
    for idx2, hotVel in np.ndenumerate(hotVelArr):
        coldVelList.append(coldVel)
        hotVelList.append(hotVel)
        ResultList.append(resArr[idx1][idx2])

tempDict = {"coldVel": coldVelList, "hotVel": hotVelList, "Result": ResultList}

df = pd.DataFrame.from_dict(tempDict)

from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import PolynomialFeatures, StandardScaler

####################################################################
# Using scikit-learn
# ==================
# * Prepare Features (X) and Label (y) using a Pre-Processing Pipeline
# * Train-Test (80-20) Split
# * Add Polynomial Features to improve ML Model

poly_features = PolynomialFeatures(degree=2, include_bias=False)

transformer1 = Pipeline(
    [
        ("poly_features", poly_features),
        ("std_scaler", StandardScaler()),
    ]
)


x_ct = ColumnTransformer(
    [
        ("transformer1", transformer1, ["coldVel", "hotVel"]),
    ],
    remainder="drop",
)

train_set, test_set = train_test_split(df, test_size=0.2, random_state=42)

X_train = x_ct.fit_transform(train_set)
X_test = x_ct.fit_transform(test_set)

y_train = train_set["Result"]
y_test = test_set["Result"]
y_train = np.ravel(y_train.T)
y_test = np.ravel(y_test.T)


####################################################
# * Define functions for:
# * Cross-Validation and Display Scores (scikit-learn)
# * Training the Model (scikit-learn)
# * Prediction on Unseen/Test Data (scikit-learn)
# * Parity Plot (Matplotlib and Seaborn)

# from pprint import pprint

# from sklearn.ensemble import RandomForestRegressor
# from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
from sklearn.model_selection import RepeatedKFold, cross_val_score
from xgboost import XGBRegressor

np.set_printoptions(precision=2)


def display_scores(scores):
    """Display scores."""
    print("\nCross-Validation Scores:", scores)
    print("Mean:%0.2f" % (scores.mean()))
    print("Std. Dev.:%0.2f" % (scores.std()))


def fit_and_predict(model):
    """Fit abd predict."""
    cv = RepeatedKFold(n_splits=5, n_repeats=3, random_state=42)
    cv_scores = cross_val_score(
        model, X_train, y_train, scoring="neg_mean_squared_error", cv=cv
    )
    rmse_scores = np.sqrt(-cv_scores)
    display_scores(rmse_scores)

    model.fit(X_train, y_train)
    train_predictions = model.predict(X_train)
    test_predictions = model.predict(X_test)
    print(train_predictions.shape[0])
    print("\n\nCoefficient Of Determination")
    print("Train Data R2 Score: %0.3f" % (r2_score(train_predictions, y_train)))
    print("Test Data R2 Score: %0.3f" % (r2_score(test_predictions, y_test)))
    print(
        "\n\nPredictions - Ground Truth (Kelvin): ", (test_predictions - y_test), "\n"
    )
    #    print("\n\nModel Parameters:")
    #    pprint(model.get_params())

    com_train_set = train_set
    com_test_set = test_set

    train_list = []
    for i in range(train_predictions.shape[0]):
        train_list.append("Train")

    test_list = []
    for i in range(test_predictions.shape[0]):
        test_list.append("Test")

    com_train_set["Result"] = train_predictions.tolist()
    com_train_set["Set"] = train_list
    com_test_set["Result"] = test_predictions.tolist()
    com_test_set["Set"] = test_list

    df_combined = pd.concat([com_train_set, com_test_set])

    df_combined.to_csv("PyFluent_Output.csv", header=True, index=False)

    fig = plt.figure(figsize=(12, 5))

    fig.add_subplot(121)
    sns.regplot(x=y_train, y=train_predictions, color="g")
    plt.title("Train Data", fontsize=16)
    plt.xlabel("Ground Truth", fontsize=12)
    plt.ylabel("Predictions", fontsize=12)

    fig.add_subplot(122)
    sns.regplot(x=y_test, y=test_predictions, color="g")
    plt.title("Unseen Data", fontsize=16)
    plt.xlabel("Ground Truth", fontsize=12)
    plt.ylabel("Predictions", fontsize=12)

    plt.tight_layout()
    plt.show()


#############################################################################
# Select the Model from Linear, Random Forest or XGBoost
# ======================================================
# * Call fit_and_predict

# model = LinearRegression()
model = XGBRegressor(
    n_estimators=100, max_depth=10, eta=0.3, subsample=0.8, random_state=42
)
# model = RandomForestRegressor(random_state=42)

fit_and_predict(model)

#############################################################################
# Show graph
# ==========

plt.show()

# %%
# .. image:: ../../_static/doe_ml_predictions_regression.png
#    :align: center
#    :alt: Regression Model Predictions

# %%
#    Regression Model Predictions

###########################################################
# 3D Visualization of Model Predictions on Train & Test Set
# =========================================================

df = pd.read_csv("PyFluent_Output.csv")

fig = px.scatter_3d(df, x="coldVel", y="hotVel", z="Result", color="Set")
fig.update_traces(marker=dict(size=4))
fig.update_layout(legend=dict(yanchor="top", y=1, xanchor="left", x=0.0))

fig.add_traces(go.Surface(z=resArr.T, x=coldVelArr, y=hotVelArr))

fig.update_layout(
    title={
        "text": "Mixing Elbow Response Surface",
        "y": 0.9,
        "x": 0.5,
        "xanchor": "center",
        "yanchor": "top",
    }
)

fig.update_layout(
    scene=dict(
        xaxis_title="Cold Inlet Vel (m/s)",
        yaxis_title="Hot Inlet Vel (m/s)",
        zaxis_title="Outlet Temperature (K)",
    ),
    width=500,
    height=500,
    margin=dict(l=80, r=80, b=80, t=80),
)

fig.show()

################################################
# TensorFlow and Keras Neural Network Regression
# ==============================================

print("TensorFlow version is:", tf.__version__)
keras.backend.clear_session()
np.random.seed(42)
tf.random.set_seed(42)

model = keras.models.Sequential(
    [
        keras.layers.Dense(
            20,
            activation="relu",
            input_shape=X_train.shape[1:],
            kernel_initializer="lecun_normal",
        ),
        keras.layers.BatchNormalization(),
        keras.layers.Dense(20, activation="relu", kernel_initializer="lecun_normal"),
        keras.layers.BatchNormalization(),
        keras.layers.Dense(20, activation="relu", kernel_initializer="lecun_normal"),
        keras.layers.BatchNormalization(),
        keras.layers.Dense(1),
    ]
)

optimizer = tf.keras.optimizers.Adam(learning_rate=0.1, beta_1=0.9, beta_2=0.999)

model.compile(loss="mean_squared_error", optimizer=optimizer)
checkpoint_cb = keras.callbacks.ModelCheckpoint(
    "my_keras_model.h5", save_best_only=True
)
early_stopping_cb = keras.callbacks.EarlyStopping(
    patience=30, restore_best_weights=True
)

model.summary()

# keras.utils.plot_model(model, show_shapes=True,) # to_file='dot_img.png', )

history = model.fit(
    X_train,
    y_train,
    epochs=250,
    validation_split=0.2,
    callbacks=[checkpoint_cb, early_stopping_cb],
)
model = keras.models.load_model("my_keras_model.h5")

print(history.params)

pd.DataFrame(history.history).plot(figsize=(8, 5))
plt.grid(True)
plt.show()

train_predictions = model.predict(X_train)
test_predictions = model.predict(X_test)
train_predictions = np.ravel(train_predictions.T)
test_predictions = np.ravel(test_predictions.T)
print(test_predictions.shape)

print("\n\nTrain R2: %0.3f" % (r2_score(train_predictions, y_train)))
print("Test R2: %0.3f" % (r2_score(test_predictions, y_test)))
print("Predictions - Ground Truth (Kelvin): ", (test_predictions - y_test))

fig = plt.figure(figsize=(12, 5))

fig.add_subplot(121)
sns.regplot(x=y_train, y=train_predictions, color="g")
plt.title("Train Data", fontsize=16)
plt.xlabel("Ground Truth", fontsize=12)
plt.ylabel("Predictions", fontsize=12)

fig.add_subplot(122)
sns.regplot(x=y_test, y=test_predictions, color="g")
plt.title("Test/Unseen Data", fontsize=16)
plt.xlabel("Ground Truth", fontsize=12)
plt.ylabel("Predictions", fontsize=12)

plt.tight_layout()


#############################################################################
# Show graph
# ==========

plt.show()

# %%
# .. image:: ../../_static/doe_ml_validation_loss.png
#    :align: center
#    :alt: Neural Network Validation Loss

# %%
#    Neural Network Validation Loss

# %%
# .. image:: ../../_static/doe_ml_predictions_neural_network.png
#    :align: center
#    :alt: Neural Network Predictions

# %%
#    Neural Network Predictions



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""A package providing job scheduler support."""

import os
import socket
import sys

from ansys.fluent.core.scheduler.load_machines import load_machines  # noqa: F401
from ansys.fluent.core.scheduler.machine_list import MachineList

_ncoresOpt = "-t%n%"
_machinesOpt = " -cnf=%machineList%"
_procSep = ":"
_machineSep = ","


def build_parallel_options(machine_list: MachineList) -> str:
    """Constructs Fluent's parallel arguments given a list of machines.

    Parameters
    ----------
    machine_list : MachineList
        List of machines obtained by calling `load_machines`.

    Notes
    -----
    When running serial no options are passed back to the caller as Fluent can
    be started without additional arguments in that case.

    If the parallel options are being built on the same machine as Fluent is run
    on, and it's local parallel, then the -cnf argument is not constructed.  On
    Windows HPC the job scheduler returns hostnames as upper case but the socket
    module may return lower case.
    """
    parOpt = ""
    if sys.platform == "win32" and "CCP_NODES" in os.environ:
        localParallel = (
            socket.gethostname().upper() == machine_list[0].host_name.upper()
        )
    else:
        localParallel = socket.gethostname() == machine_list[0].host_name
    if machine_list.num_machines == 1 and localParallel:
        if machine_list.number_of_cores > 1:
            parOpt = _ncoresOpt.replace("%n%", str(machine_list.number_of_cores))
    else:
        parOpt = _ncoresOpt.replace("%n%", str(machine_list.number_of_cores))
        cnfList = (
            machine_list[0].host_name + _procSep + str(machine_list[0].number_of_cores)
        )
        for m in range(1, len(machine_list)):
            cnfList += (
                _machineSep
                + machine_list[m].host_name
                + _procSep
                + str(machine_list[m].number_of_cores)
            )
        parOpt += _machinesOpt.replace("%machineList%", cnfList)
    return parOpt
